"""Utility functions for finders."""
import time
import abc

from graphite.node import BranchNode, LeafNode  # noqa
from graphite.util import is_pattern
from graphite.intervals import Interval


class FindQuery(object):
    def __init__(self, pattern, startTime, endTime,
                 local=False, headers=None, leaves_only=None):
        self.pattern = pattern
        self.startTime = startTime
        self.endTime = endTime
        self.isExact = is_pattern(pattern)
        self.interval = Interval(
            float('-inf') if startTime is None else startTime,
            float('inf') if endTime is None else endTime)
        self.local = local
        self.headers = headers
        self.leaves_only = leaves_only

    def __repr__(self):
        if self.startTime is None:
            startString = '*'
        else:
            startString = time.ctime(self.startTime)

        if self.endTime is None:
            endString = '*'
        else:
            endString = time.ctime(self.endTime)

        return '<FindQuery: %s from %s until %s>' % (
            self.pattern, startString, endString)


class BaseFinder(object):
    __metaclass__ = abc.ABCMeta

    # Set to False if this is a remote finder.
    local = True
    # set to True if this finder shouldn't be used
    disabled = False
    # set to True if this finder supports seriesByTag
    tags = False

    def __init__(self):
        """Initialize the finder."""

    @abc.abstractmethod
    def find_nodes(self, query):
        """Get the list of nodes matching a query.

        Args:
          graphite.storage.FindQuery: the query to run.

        Returns:
          generator of Node
        """

    def get_index(self, requestContext):
        """Get a list of all series

        Args:
          requestContext

        Returns:
          list of series
        """
        query = FindQuery(
            '**', None, None,
            local=requestContext.get('localOnly'),
            headers=requestContext.get('forwardHeaders'),
            leaves_only=True,
        )

        return sorted([node.path for node in self.find_nodes(query) if node.is_leaf])

    # The methods below are fully optional and BaseFinder provides
    # a default implementation. They can be re-implemented by finders
    # that could provide a more efficient way of doing it.
    #
    # The API isn't fully finalized yet and can be subject to change
    # until it's documented.

    @classmethod
    def factory(cls):
        return [cls()]

    def find_multi(self, queries):
        """Executes multiple find queries.

        Works like multiple find_node(), this gives the ability for
        finders to parallelize the work.

        Returns:
          generator of (Node, query)
        """
        for query in queries:
            for node in self.find_nodes(query):
                yield (node, query)

    def fetch(self, patterns, start_time, end_time, now=None, requestContext=None):
        """Fetch multiple patterns at once.

        This method is used to fetch multiple patterns at once, this
        allows alternate finders to do batching on their side when they can.

        Returns:
          an iterable of
          {
            'pathExpression': pattern,
            'path': node.path,
            'time_info': time_info,
            'values': values,
          }
        """
        requestContext = requestContext or {}

        queries = [
            FindQuery(
                pattern, start_time, end_time,
                local=requestContext.get('localOnly'),
                headers=requestContext.get('forwardHeaders'),
                leaves_only=True,
            )
            for pattern in patterns
        ]

        results = []

        for node, query in self.find_multi(queries):
            if not isinstance(node, LeafNode):
                continue

            result = node.fetch(
                start_time, end_time,
                now=now, requestContext=requestContext
            )

            if result is None:
                continue

            time_info, values = result

            results.append({
                'pathExpression': query.pattern,
                'path': node.path,
                'name': node.path,
                'time_info': time_info,
                'values': values,
            })

        return results

    def auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
        return []

    def auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
        return []



from django.http import HttpResponseBadRequest
from graphite.logger import log


class NormalizeEmptyResultError(Exception):
    # throw error for normalize() when empty
    pass


class InputParameterError(ValueError):

    def __init__(self, *args, **kwargs):
        super(InputParameterError, self).__init__(*args, **kwargs)
        self.context = {}

    def setSourceIdHeaders(self, newHeaders):
        headers = self.context.get('sourceIdHeaders', {})
        headers.update(newHeaders)
        self.context['sourceIdHeaders'] = headers

    @property
    def sourceIdHeaders(self):
        sourceIdHeaders = self.context.get('sourceIdHeaders', {})
        headers = list(sourceIdHeaders.keys())
        headers.sort()
        source = ''

        for name in headers:
            if source:
                source += ', '
            source += '{name}: {value}'.format(
                name=name,
                value=sourceIdHeaders[name])

        return source

    def setTargets(self, targets):
        self.context['targets'] = targets

    @property
    def targets(self):
        return ', '.join(self.context.get('targets', []))

    def setFunction(self, name, args, kwargs):
        self.context['function'] = {
            'name': name,
            'args': args,
            'kwargs': kwargs,
        }

    @property
    def function(self):
        func = self.context.get('function', None)
        if not func:
            return ''

        funcName = func.get('name', '')
        if not funcName:
            return ''

        kwargs = func.get('kwargs', {})
        kwargKeys = list(kwargs.keys())

        # keep kwargs sorted in message, for consistency and testability
        kwargKeys.sort()

        # generate string of args and kwargs
        args = ', '.join(
            argList
            for argList in [
                ', '.join(repr(arg) for arg in func.get('args', [])),
                ', '.join('{k}={v}'.format(k=str(k), v=repr(kwargs[k])) for k in kwargKeys),
            ] if argList
        )

        return '{func}({args})'.format(func=funcName, args=args)

    def __str__(self):
        msg = 'Invalid parameters ({msg})'.format(msg=str(super(InputParameterError, self).__str__()))
        targets = self.targets
        if targets:
            msg += '; targets: "{targets}"'.format(targets=targets)

        source = self.sourceIdHeaders
        if source:
            msg += '; source: "{source}"'.format(source=source)

        # needs to be last because the string "args" may potentially be thousands
        # of chars long after expanding the globbing patterns
        func = self.function
        if func:
            msg += '; func: "{func}"'.format(func=func)

        return msg


# decorator which turns InputParameterExceptions into Django's HttpResponseBadRequest
def handleInputParameterError(f):
    def new_f(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except InputParameterError as e:
            msgStr = str(e)
            log.warning('%s', msgStr)
            return HttpResponseBadRequest(msgStr)

    return new_f



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import itertools
import math
import re
from datetime import datetime, timedelta

from six.moves import range, zip
from six.moves.urllib.parse import unquote_plus
from six.moves.configparser import SafeConfigParser
from django.conf import settings
import pytz
import six

from graphite.render.datalib import TimeSeries
from graphite.util import json, BytesIO

try:
    import cairocffi as cairo
except ImportError:
    import cairo

INFINITY = float('inf')

colorAliases = {
  'black' : (0,0,0),
  'white' : (255,255,255),
  'blue' : (100,100,255),
  'green' : (0,200,0),
  'red' : (200,00,50),
  'yellow' : (255,255,0),
  'orange' : (255, 165, 0),
  'purple' : (200,100,255),
  'brown' : (150,100,50),
  'cyan' : (0,255,255),
  'aqua' : (0,150,150),
  'gray' : (175,175,175),
  'grey' : (175,175,175),
  'magenta' : (255,0,255),
  'pink' : (255,100,100),
  'gold' : (200,200,0),
  'rose' : (200,150,200),
  'darkblue' : (0,0,255),
  'darkgreen' : (0,255,0),
  'darkred' : (255,0,0),
  'darkgray' : (111,111,111),
  'darkgrey' : (111,111,111),
}

# This gets overridden by graphTemplates.conf
defaultGraphOptions = dict(
  background='white',
  foreground='black',
  majorline='rose',
  minorline='grey',
  linecolors='blue,green,red,purple,brown,yellow,aqua,grey,magenta,pink,gold,rose',
  fontname='Sans',
  fontsize=10,
  fontbold='false',
  fontitalic='false',
)

# X-axis configurations (copied from rrdtool, this technique is evil & ugly but effective)
SEC = 1
MIN = 60
HOUR = MIN * 60
DAY = HOUR * 24
WEEK = DAY * 7
MONTH = DAY * 31
YEAR = DAY * 365

# Set a flag to indicate whether the '%l' option can be used safely.
# On Windows, in particular the %l option in strftime is not supported.
# '%l' can also fail silently in Linux.
# (It is not one of the documented Python formatters).
try:
    if datetime.now().strftime("%a %l%p"):
        percent_l_supported = True
    else:
        percent_l_supported = False
except ValueError:
    percent_l_supported = False

DATE_FORMAT = settings.DATE_FORMAT

xAxisConfigs = (
  dict(seconds=0.00,  minorGridUnit=SEC,  minorGridStep=5,  majorGridUnit=MIN,  majorGridStep=1,  labelUnit=SEC,  labelStep=5,  format="%H:%M:%S", maxInterval=10*MIN),
  dict(seconds=0.07,  minorGridUnit=SEC,  minorGridStep=10, majorGridUnit=MIN,  majorGridStep=1,  labelUnit=SEC,  labelStep=10, format="%H:%M:%S", maxInterval=20*MIN),
  dict(seconds=0.14,  minorGridUnit=SEC,  minorGridStep=15, majorGridUnit=MIN,  majorGridStep=1,  labelUnit=SEC,  labelStep=15, format="%H:%M:%S", maxInterval=30*MIN),
  dict(seconds=0.27,  minorGridUnit=SEC,  minorGridStep=30, majorGridUnit=MIN,  majorGridStep=2,  labelUnit=MIN,  labelStep=1,  format="%H:%M", maxInterval=2*HOUR),
  dict(seconds=0.5,   minorGridUnit=MIN,  minorGridStep=1,  majorGridUnit=MIN,  majorGridStep=2,  labelUnit=MIN,  labelStep=1,  format="%H:%M", maxInterval=2*HOUR),
  dict(seconds=1.2,   minorGridUnit=MIN,  minorGridStep=1,  majorGridUnit=MIN,  majorGridStep=4,  labelUnit=MIN,  labelStep=2,  format="%H:%M", maxInterval=3*HOUR),
  dict(seconds=2,     minorGridUnit=MIN,  minorGridStep=1,  majorGridUnit=MIN,  majorGridStep=10, labelUnit=MIN,  labelStep=5,  format="%H:%M", maxInterval=6*HOUR),
  dict(seconds=5,     minorGridUnit=MIN,  minorGridStep=2,  majorGridUnit=MIN,  majorGridStep=10, labelUnit=MIN,  labelStep=10, format="%H:%M", maxInterval=12*HOUR),
  dict(seconds=10,    minorGridUnit=MIN,  minorGridStep=5,  majorGridUnit=MIN,  majorGridStep=20, labelUnit=MIN,  labelStep=20, format="%H:%M", maxInterval=1*DAY),
  dict(seconds=30,    minorGridUnit=MIN,  minorGridStep=10, majorGridUnit=HOUR, majorGridStep=1,  labelUnit=HOUR, labelStep=1,  format="%H:%M", maxInterval=2*DAY),
  dict(seconds=60,    minorGridUnit=MIN,  minorGridStep=30, majorGridUnit=HOUR, majorGridStep=2,  labelUnit=HOUR, labelStep=2,  format="%H:%M", maxInterval=2*DAY),
  dict(seconds=100,   minorGridUnit=HOUR, minorGridStep=2,  majorGridUnit=HOUR, majorGridStep=4,  labelUnit=HOUR, labelStep=4,  format="%a %H:%M", maxInterval=6*DAY),
  dict(seconds=255,   minorGridUnit=HOUR, minorGridStep=6,  majorGridUnit=HOUR, majorGridStep=12, labelUnit=HOUR, labelStep=12, format=DATE_FORMAT + " %H:%M", maxInterval=10*DAY),
  dict(seconds=600,   minorGridUnit=HOUR, minorGridStep=6,  majorGridUnit=DAY,  majorGridStep=1,  labelUnit=DAY,  labelStep=1,  format=DATE_FORMAT, maxInterval=14*DAY),
  dict(seconds=1000,   minorGridUnit=HOUR, minorGridStep=12, majorGridUnit=DAY,  majorGridStep=1,  labelUnit=DAY,  labelStep=1,  format=DATE_FORMAT, maxInterval=365*DAY),
  dict(seconds=2000,  minorGridUnit=DAY,  minorGridStep=1,  majorGridUnit=DAY,  majorGridStep=2,  labelUnit=DAY,  labelStep=2,  format=DATE_FORMAT, maxInterval=365*DAY),
  dict(seconds=4000,  minorGridUnit=DAY,  minorGridStep=2,  majorGridUnit=DAY,  majorGridStep=4,  labelUnit=DAY,  labelStep=4,  format=DATE_FORMAT, maxInterval=365*DAY),
  dict(seconds=8000,  minorGridUnit=DAY,  minorGridStep=3.5,majorGridUnit=DAY,  majorGridStep=7,  labelUnit=DAY,  labelStep=7,  format=DATE_FORMAT, maxInterval=365*DAY),
  dict(seconds=16000, minorGridUnit=DAY,  minorGridStep=7,  majorGridUnit=DAY,  majorGridStep=14, labelUnit=DAY,  labelStep=14, format=DATE_FORMAT, maxInterval=365*DAY),
  dict(seconds=32000, minorGridUnit=DAY,  minorGridStep=15, majorGridUnit=DAY,  majorGridStep=30, labelUnit=DAY,  labelStep=30, format=DATE_FORMAT, maxInterval=365*DAY),
  dict(seconds=64000, minorGridUnit=DAY,  minorGridStep=30, majorGridUnit=DAY,  majorGridStep=60, labelUnit=DAY,  labelStep=60, format=DATE_FORMAT + " %Y"),
  dict(seconds=100000,minorGridUnit=DAY,  minorGridStep=60, majorGridUnit=DAY,  majorGridStep=120,labelUnit=DAY,  labelStep=120, format=DATE_FORMAT + " %Y"),
  dict(seconds=120000,minorGridUnit=DAY,  minorGridStep=120,majorGridUnit=DAY,  majorGridStep=240,labelUnit=DAY,  labelStep=240, format=DATE_FORMAT + " %Y"),
)

UnitSystems = {
  'binary': (
    ('Pi', 1024.0**5),
    ('Ti', 1024.0**4),
    ('Gi', 1024.0**3),
    ('Mi', 1024.0**2),
    ('Ki', 1024.0   )),
  'si': (
    ('P', 1000.0**5),
    ('T', 1000.0**4),
    ('G', 1000.0**3),
    ('M', 1000.0**2),
    ('k', 1000.0   )),
  'sec': (
    ('Y', 60*60*24*365),
    ('M', 60*60*24*30),
    ('D', 60*60*24),
    ('H', 60*60),
    ('m', 60)),
  'msec': (
    ('Y', 60*60*24*365*1000),
    ('M', 60*60*24*30*1000),
    ('D', 60*60*24*1000),
    ('H', 60*60*1000),
    ('m', 60*1000),
    ('s', 1000)),
  'none' : [],
}


# We accept values fractionally outside of nominal limits, so that
# rounding errors don't cause weird effects. Since our goal is to
# create plots, and the maximum resolution of the plots is likely to
# be less than 10000 pixels, errors smaller than this size shouldn't
# create any visible effects.
EPSILON = 0.0001


class GraphError(Exception):
  pass


class _AxisTics:
  def __init__(self, minValue, maxValue, unitSystem=None):
    self.minValue = self.checkFinite(minValue, "data value")
    self.minValueSource = 'data'
    self.maxValue = self.checkFinite(maxValue, "data value")
    self.maxValueSource = 'data'
    self.unitSystem = unitSystem

  @staticmethod
  def checkFinite(value, name='value'):
    """Check that value is a finite number.

    If it is, return it. If not, raise GraphError describing the
    problem, using name in the error message.
    """

    if math.isnan(value):
      raise GraphError('Encountered NaN %s' % (name,))
    elif math.isinf(value):
      raise GraphError('Encountered infinite %s' % (name,))
    return value

  @staticmethod
  def chooseDelta(x):
    """Choose a reasonable axis range given that one limit is x.

    Given that end of the axis range (i.e., minValue or maxValue) is
    x, choose a reasonable distance to the other limit.
    """

    if abs(x) < 1.0e-9:
      return 1.0
    else:
      return 0.1 * abs(x)

  def reconcileLimits(self):
    """If self.minValue is not less than self.maxValue, fix the problem.

    If self.minValue is not less than self.maxValue, adjust
    self.minValue and/or self.maxValue (depending on which was not
    specified explicitly by the user) to make self.minValue <
    self.maxValue. If the user specified both limits explicitly, then
    raise GraphError.
    """

    if self.minValue < self.maxValue:
      # The limits are already OK.
      return

    minFixed = (self.minValueSource in ['min'])
    maxFixed = (self.maxValueSource in ['max', 'limit'])

    if minFixed and maxFixed:
      raise GraphError('The %s must be less than the %s' %
                       (self.minValueSource, self.maxValueSource))
    elif minFixed:
      self.maxValue = self.minValue + self.chooseDelta(self.minValue)
    elif maxFixed:
      self.minValue = self.maxValue - self.chooseDelta(self.maxValue)
    else:
      delta = self.chooseDelta(max(abs(self.minValue), abs(self.maxValue)))
      average = (self.minValue + self.maxValue) / 2.0
      self.minValue = average - delta
      self.maxValue = average + delta

  def applySettings(self, axisMin=None, axisMax=None, axisLimit=None):
    """Apply the specified settings to this axis.

    Set self.minValue, self.minValueSource, self.maxValue,
    self.maxValueSource, and self.axisLimit reasonably based on the
    parameters provided.

    Arguments:

    axisMin -- a finite number, or None to choose a round minimum
        limit that includes all of the data.

    axisMax -- a finite number, 'max' to use the maximum value
        contained in the data, or None to choose a round maximum limit
        that includes all of the data.

    axisLimit -- a finite number to use as an upper limit on maxValue,
        or None to impose no upper limit.

    """

    if axisMin is not None and not math.isnan(axisMin):
      self.minValueSource = 'min'
      self.minValue = self.checkFinite(axisMin, 'axis min')

    if axisMax == 'max':
      self.maxValueSource = 'extremum'
    elif axisMax is not None and not math.isnan(axisMax):
      self.maxValueSource = 'max'
      self.maxValue = self.checkFinite(axisMax, 'axis max')

    if axisLimit is None or math.isnan(axisLimit):
      self.axisLimit = None
    elif axisLimit < self.maxValue:
      self.maxValue = self.checkFinite(axisLimit, 'axis limit')
      self.maxValueSource = 'limit'
      # The limit has already been imposed, so there is no need to
      # remember it:
      self.axisLimit = None
    elif math.isinf(axisLimit):
      # It must be positive infinity, which is the same as no limit:
      self.axisLimit = None
    else:
      # We still need to remember axisLimit to avoid rounding top to
      # a value larger than axisLimit:
      self.axisLimit = axisLimit

    self.reconcileLimits()

  def makeLabel(self, value):
    """Create a label for the specified value.

    Create a label string containing the value and its units (if any),
    based on the values of self.step, self.span, and self.unitSystem.

    """

    value, prefix = format_units(value, self.step, system=self.unitSystem)
    span, spanPrefix = format_units(self.span, self.step, system=self.unitSystem)
    if prefix:
      prefix += " "
    if value < 0.1:
      return "%g %s" % (float(value), prefix)
    elif value < 1.0:
      return "%.2f %s" % (float(value), prefix)
    if (span is not None and span > 10) or spanPrefix != prefix:
      if type(value) is float:
        return "%.1f %s" % (value, prefix)
      else:
        return "%d %s" % (int(value), prefix)
    elif span is not None and span > 3:
      return "%.1f %s" % (float(value), prefix)
    elif span is not None and span > 0.1:
      return "%.2f %s" % (float(value), prefix)
    else:
      return "%g %s" % (float(value), prefix)


class _LinearAxisTics(_AxisTics):
  """Axis ticmarks with uniform spacing."""

  def __init__(self, minValue, maxValue, unitSystem=None):
    _AxisTics.__init__(self, minValue, maxValue, unitSystem=unitSystem)
    self.step = None
    self.span = None
    self.binary = None

  def setStep(self, step):
    """Set the size of steps between ticmarks."""

    self.step = self.checkFinite(float(step), 'axis step')

  def generateSteps(self, minStep):
    """Generate allowed steps with step >= minStep in increasing order."""

    self.checkFinite(minStep)

    if self.binary:
      base = 2.0
      mantissas = [1.0]
      exponent = math.floor(math.log(minStep, 2) - EPSILON)
    else:
      base = 10.0
      mantissas = [1.0, 2.0, 5.0]
      exponent = math.floor(math.log10(minStep) - EPSILON)

    while True:
      multiplier = base ** exponent
      for mantissa in mantissas:
        value = mantissa * multiplier
        if value >= minStep * (1.0 - EPSILON):
          yield value
      exponent += 1

  def computeSlop(self, step, divisor):
    """Compute the slop that would result from step and divisor.

    Return the slop, or None if this combination can't cover the full
    range. See chooseStep() for the definition of "slop".

    """

    bottom = step * math.floor(self.minValue / float(step) + EPSILON)
    top = bottom + step * divisor

    if top >= self.maxValue - EPSILON * step:
      return max(top - self.maxValue, self.minValue - bottom)
    else:
      return None

  def chooseStep(self, divisors=None, binary=False):
    """Choose a nice, pretty size for the steps between axis labels.

    Our main constraint is that the number of divisions must be taken
    from the divisors list. We pick a number of divisions and a step
    size that minimizes the amount of whitespace ("slop") that would
    need to be included outside of the range [self.minValue,
    self.maxValue] if we were to push out the axis values to the next
    larger multiples of the step size.

    The minimum step that could possibly cover the variance satisfies

        minStep * max(divisors) >= variance

    or

        minStep = variance / max(divisors)

    It's not necessarily possible to cover the variance with a step
    that size, but we know that any smaller step definitely *cannot*
    cover it. So we can start there.

    For a sufficiently large step size, it is definitely possible to
    cover the variance, but at some point the slop will start growing.
    Let's define the slop to be

        slop = max(minValue - bottom, top - maxValue)

    Then for a given, step size, we know that

        slop >= (1/2) * (step * min(divisors) - variance)

    (the factor of 1/2 is for the best-case scenario that the slop is
    distributed equally on the two sides of the range). So suppose we
    already have a choice that yields bestSlop. Then there is no need
    to choose steps so large that the slop is guaranteed to be larger
    than bestSlop. Therefore, the maximum step size that we need to
    consider is

        maxStep = (2 * bestSlop + variance) / min(divisors)

    """

    self.binary = binary
    if divisors is None:
      divisors = [4,5,6]
    else:
      for divisor in divisors:
        self.checkFinite(divisor, 'divisor')
        if divisor < 1:
          raise GraphError('Divisors must be greater than or equal to one')

    if self.minValue == self.maxValue:
      if self.minValue == 0.0:
        self.maxValue = 1.0
      elif self.minValue < 0.0:
        self.minValue *= 1.1
        self.maxValue *= 0.9
      else:
        self.minValue *= 0.9
        self.maxValue *= 1.1

    variance = self.maxValue - self.minValue

    bestSlop = None
    bestStep = None
    for step in self.generateSteps(variance / float(max(divisors))):
      if bestSlop is not None and step * min(divisors) >= 2 * bestSlop + variance:
        break
      for divisor in divisors:
        slop = self.computeSlop(step, divisor)
        if slop is not None and (bestSlop is None or slop < bestSlop):
          bestSlop = slop
          bestStep = step

    self.step = bestStep

  def chooseLimits(self):
    if self.minValueSource == 'data':
      # Start labels at the greatest multiple of step <= minValue:
      self.bottom = self.step * math.floor(self.minValue / self.step + EPSILON)
    else:
      self.bottom = self.minValue

    if self.maxValueSource == 'data':
      # Extend the top of our graph to the lowest step multiple >= maxValue:
      self.top = self.step * math.ceil(self.maxValue / self.step - EPSILON)
      # ...but never exceed a user-specified limit:
      if self.axisLimit is not None and self.top > self.axisLimit + EPSILON * self.step:
        self.top = self.axisLimit
    else:
      self.top = self.maxValue

    self.span = self.top - self.bottom

    if self.span == 0:
      self.top += 1
      self.span += 1

  def getLabelValues(self):
    if self.step <= 0.0:
      raise GraphError('The step size must be positive')
    if self.span > 1000.0 * self.step:
      # This is insane. Pick something that won't cause trouble:
      self.chooseStep()

    values = []

    start = self.step * math.ceil(self.bottom / self.step - EPSILON)
    i = 0
    while True:
      value = start + i * self.step
      if value > self.top + EPSILON * self.step:
        break
      values.append(value)
      i += 1

    return values


class _LogAxisTics(_AxisTics):
  def __init__(self, minValue, maxValue, unitSystem=None, base=10.0):
    _AxisTics.__init__(self, minValue, maxValue, unitSystem=unitSystem)
    if base <= 1.0:
      raise GraphError('Logarithmic base must be greater than one')
    self.base = self.checkFinite(base, 'log base')
    self.step = None
    self.span = None

  def setStep(self, step):
    # step is ignored for Logarithmic tics:
    self.step = None

  def chooseStep(self, divisors=None, binary=False):
    # step is ignored for Logarithmic tics:
    self.step = None

  def chooseLimits(self):
    if self.minValue <= 0:
      raise GraphError('Logarithmic scale specified with a dataset with a '
                       'minimum value less than or equal to zero')
    self.bottom = math.pow(self.base, math.floor(math.log(self.minValue, self.base)))
    self.top = math.pow(self.base, math.ceil(math.log(self.maxValue, self.base)))

    self.span = self.top - self.bottom

    if self.span == 0:
      self.top *= self.base
      self.span = self.top - self.bottom

  def getLabelValues(self):
    values = []

    value = math.pow(self.base, math.ceil(math.log(self.bottom, self.base) - EPSILON))
    while value < self.top * (1.0 + EPSILON):
       values.append(value)
       value *= self.base

    return values


class Graph:
  customizable = ('width','height','margin','bgcolor','fgcolor',
                 'fontName','fontSize','fontBold','fontItalic',
                 'colorList','template','yAxisSide','outputFormat')

  def __init__(self,**params):
    self.params = params
    self.data = params['data']
    self.dataLeft = []
    self.dataRight = []
    self.secondYAxis = False
    self.width = int( params.get('width',200) )
    self.height = int( params.get('height',200) )
    self.margin = int( params.get('margin',10) )
    self.userTimeZone = params.get('tz')
    self.logBase = params.get('logBase', None)
    self.minorY = int(params.get('minorY', 1))

    if self.logBase:
      if self.logBase == 'e':
        self.logBase = math.e
      elif self.logBase < 1:
        self.logBase = None
        params['logBase'] = None
      else:
        self.logBase = float(self.logBase)

    if self.margin < 0:
      self.margin = 10

    self.setupCairo( params.get('outputFormat','png').lower() )

    self.area = {
      'xmin' : self.margin + 10, # Need extra room when the time is near the left edge
      'xmax' : self.width - self.margin,
      'ymin' : self.margin,
      'ymax' : self.height - self.margin,
    }

    self.loadTemplate( params.get('template','default') )

    opts = self.ctx.get_font_options()
    opts.set_antialias( cairo.ANTIALIAS_NONE )
    self.ctx.set_font_options( opts )

    self.foregroundColor = params.get('fgcolor',self.defaultForeground)
    self.backgroundColor = params.get('bgcolor',self.defaultBackground)
    self.setColor( self.backgroundColor )
    self.drawRectangle( 0, 0, self.width, self.height )

    if 'colorList' in params:
      colorList = unquote_plus( str(params['colorList']) ).split(',')
    else:
      colorList = self.defaultColorList
    self.colors = itertools.cycle( colorList )

    self.drawGraph(**params)

  def setupCairo(self,outputFormat='png'):
    self.outputFormat = outputFormat
    if outputFormat == 'png':
      self.surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.width, self.height)
    elif outputFormat == 'svg':
      self.surfaceData = BytesIO()
      self.surface = cairo.SVGSurface(self.surfaceData, self.width, self.height)
    elif outputFormat == 'pdf':
      self.surfaceData = BytesIO()
      self.surface = cairo.PDFSurface(self.surfaceData, self.width, self.height)
      res_x, res_y = self.surface.get_fallback_resolution()
      self.width = float(self.width / res_x) * 72
      self.height = float(self.height / res_y) * 72
      self.surface.set_size(self.width, self.height)
    self.ctx = cairo.Context(self.surface)

  def setColor(self, value, alpha=1.0, forceAlpha=False):
    if type(value) is tuple and len(value) == 3:
      r,g,b = value
    elif value in colorAliases:
      r,g,b = colorAliases[value]
    elif isinstance(value, six.string_types) and len(value) >= 6:
      s = value
      if s[0] == '#': s = s[1:]
      if s[0:3] == '%23': s = s[3:]
      r,g,b = ( int(s[0:2],base=16), int(s[2:4],base=16), int(s[4:6],base=16) )
      if len(s) == 8 and not forceAlpha:
        alpha = float( int(s[6:8],base=16) ) / 255.0
    elif isinstance(value, int) and len(str(value)) == 6:
      s = str(value)
      r,g,b = ( int(s[0:2],base=16), int(s[2:4],base=16), int(s[4:6],base=16) )
    else:
      raise ValueError("Must specify an RGB 3-tuple, an html color string, or a known color alias!")
    r,g,b = [float(c) / 255.0 for c in (r,g,b)]
    self.ctx.set_source_rgba(r,g,b,alpha)

  def setFont(self, **params):
    p = self.defaultFontParams.copy()
    p.update(params)
    self.ctx.select_font_face(p['name'], p['italic'], p['bold'])
    self.ctx.set_font_size( float(p['size']) )

  def getExtents(self,text=None,fontOptions={}):
    if fontOptions:
      self.setFont(**fontOptions)
    F = self.ctx.font_extents()
    extents = { 'maxHeight' : F[2], 'maxAscent' : F[0], 'maxDescent' : F[1] }
    if text:
      T = self.ctx.text_extents(text)
      extents['width'] = T[4]
      extents['height'] = T[3]
    return extents

  def drawRectangle(self, x, y, w, h, fill=True, dash=False):
    if not fill:
      o = self.ctx.get_line_width() / 2.0 #offset for borders so they are drawn as lines would be
      x += o
      y += o
      w -= o
      h -= o
    self.ctx.rectangle(x,y,w,h)
    if fill:
      self.ctx.fill()
    else:
      if dash:
        self.ctx.set_dash(dash,1)
      else:
        self.ctx.set_dash([],0)
      self.ctx.stroke()

  def drawText(self,text,x,y,font={},color={},align='left',valign='top',border=False,rotate=0):
    if font: self.setFont(**font)
    if color: self.setColor(**color)
    extents = self.getExtents(text)
    angle = math.radians(rotate)
    origMatrix = self.ctx.get_matrix()

    horizontal = {
      'left' : 0,
      'center' : extents['width'] / 2,
      'right' : extents['width'],
    }[align.lower()]
    vertical = {
      'top' : extents['maxAscent'],
      'middle' : extents['maxHeight'] / 2 - extents['maxDescent'],
      'bottom' : -extents['maxDescent'],
      'baseline' : 0,
    }[valign.lower()]

    self.ctx.move_to(x,y)
    self.ctx.rel_move_to( math.sin(angle) * -vertical, math.cos(angle) * vertical)
    self.ctx.rotate(angle)
    self.ctx.rel_move_to( -horizontal, 0 )
    bx, by = self.ctx.get_current_point()
    by -= extents['maxAscent']
    self.ctx.text_path(text)
    self.ctx.fill()
    if border:
      self.drawRectangle(bx, by, extents['width'], extents['maxHeight'], fill=False)
    else:
      self.ctx.set_matrix(origMatrix)

  def drawTitle(self,text):
    self.encodeHeader('title')

    y = self.area['ymin']
    x = self.width / 2
    lineHeight = self.getExtents()['maxHeight']
    for line in text.split('\n'):
      self.drawText(line, x, y, align='center')
      y += lineHeight
    if self.params.get('yAxisSide') == 'right':
      self.area['ymin'] = y
    else:
      self.area['ymin'] = y + self.margin

  def drawLegend(self, elements, unique=False): #elements is [ (name,color,rightSide), (name,color,rightSide), ... ]
    self.encodeHeader('legend')

    if unique:
      # remove duplicate names
      namesSeen = []
      newElements = []
      for e in elements:
        if e[0] not in namesSeen:
          namesSeen.append(e[0])
          newElements.append(e)
      elements = newElements

    # Check if there's enough room to use two columns.
    rightSideLabels = False
    padding = 5
    longestName = sorted([e[0] for e in elements],key=len)[-1]
    testSizeName = longestName + " " + longestName # Double it to check if there's enough room for 2 columns
    testExt = self.getExtents(testSizeName)
    testBoxSize = testExt['maxHeight'] - 1
    testWidth = testExt['width'] + 2 * (testBoxSize + padding)
    if testWidth + 50 < self.width:
      rightSideLabels = True

    if(self.secondYAxis and rightSideLabels):
      extents = self.getExtents(longestName)
      padding = 5
      boxSize = extents['maxHeight'] - 1
      lineHeight = extents['maxHeight'] + 1
      labelWidth = extents['width'] + 2 * (boxSize + padding)
      columns = max(1, math.floor( (self.width - self.area['xmin']) / labelWidth ))
      numRight = len([name for (name,color,rightSide) in elements if rightSide])
      numberOfLines = max(len(elements) - numRight, numRight)
      columns = math.floor(columns / 2.0)
      if columns < 1: columns = 1
      legendHeight = (max(1, (numberOfLines / columns)) * lineHeight) + padding
      self.area['ymax'] -= legendHeight #scoot the drawing area up to fit the legend
      self.ctx.set_line_width(1.0)
      x = self.area['xmin']
      y = self.area['ymax'] + (2 * padding)
      n = 0
      xRight = self.area['xmax'] - self.area['xmin']
      yRight = y
      nRight = 0
      for (name,color,rightSide) in elements:
        self.setColor( color )
        if rightSide:
          nRight += 1
          self.drawRectangle(xRight - padding,yRight,boxSize,boxSize)
          self.setColor( 'darkgrey' )
          self.drawRectangle(xRight - padding,yRight,boxSize,boxSize,fill=False)
          self.setColor( self.foregroundColor )
          self.drawText(name, xRight - boxSize, yRight, align='right')
          xRight -= labelWidth
          if nRight % columns == 0:
            xRight = self.area['xmax'] - self.area['xmin']
            yRight += lineHeight
        else:
          n += 1
          self.drawRectangle(x,y,boxSize,boxSize)
          self.setColor( 'darkgrey' )
          self.drawRectangle(x,y,boxSize,boxSize,fill=False)
          self.setColor( self.foregroundColor )
          self.drawText(name, x + boxSize + padding, y, align='left')
          x += labelWidth
          if n % columns == 0:
            x = self.area['xmin']
            y += lineHeight
    else:
      extents = self.getExtents(longestName)
      boxSize = extents['maxHeight'] - 1
      lineHeight = extents['maxHeight'] + 1
      labelWidth = extents['width'] + 2 * (boxSize + padding)
      columns = math.floor( self.width / labelWidth )
      if columns < 1: columns = 1
      numberOfLines = math.ceil( float(len(elements)) / columns )
      legendHeight = (numberOfLines * lineHeight) + padding
      self.area['ymax'] -= legendHeight #scoot the drawing area up to fit the legend
      self.ctx.set_line_width(1.0)
      x = self.area['xmin']
      y = self.area['ymax'] + (2 * padding)
      for i,(name,color,rightSide) in enumerate(elements):
        if rightSide:
          self.setColor( color )
          self.drawRectangle(x + labelWidth + padding,y,boxSize,boxSize)
          self.setColor( 'darkgrey' )
          self.drawRectangle(x + labelWidth + padding,y,boxSize,boxSize,fill=False)
          self.setColor( self.foregroundColor )
          self.drawText(name, x + labelWidth, y, align='right')
          x += labelWidth
        else:
          self.setColor( color )
          self.drawRectangle(x,y,boxSize,boxSize)
          self.setColor( 'darkgrey' )
          self.drawRectangle(x,y,boxSize,boxSize,fill=False)
          self.setColor( self.foregroundColor )
          self.drawText(name, x + boxSize + padding, y, align='left')
          x += labelWidth
        if (i + 1) % columns == 0:
          x = self.area['xmin']
          y += lineHeight

  def encodeHeader(self,text):
    """
    Puts some metadata in the generated svg xml that is not inside the frame.
    This can be used to manipulate the svg later on with a framework like d3.
    """
    if self.outputFormat != 'svg':
      return
    self.ctx.save()
    self.setColor( self.backgroundColor )
    self.ctx.move_to(-88,-88) # identifier
    for i, char in enumerate(text):
      self.ctx.line_to(-ord(char), -i-1)
    self.ctx.stroke()
    self.ctx.restore()

  def loadTemplate(self,template):
    conf = SafeConfigParser()
    if conf.read(settings.GRAPHTEMPLATES_CONF):
      defaults = defaultGraphOptions
      # If a graphTemplates.conf exists, read in
      # the values from it, but make sure that
      # all of the default values properly exist
      defaults.update(dict(conf.items('default')))
      if template in conf.sections():
        opts = dict( conf.items(template) )
      else:
        opts = defaults
    else:
      opts = defaults = defaultGraphOptions

    self.defaultBackground = opts.get('background', defaults['background'])
    self.defaultForeground = opts.get('foreground', defaults['foreground'])
    self.defaultMajorGridLineColor = opts.get('majorline', defaults['majorline'])
    self.defaultMinorGridLineColor = opts.get('minorline', defaults['minorline'])
    self.defaultColorList = [c.strip() for c in opts.get('linecolors', defaults['linecolors']).split(',')]
    fontName = opts.get('fontname', defaults['fontname'])
    fontSize = float( opts.get('fontsize', defaults['fontsize']) )
    fontBold = opts.get('fontbold', defaults['fontbold']).lower() == 'true'
    fontItalic = opts.get('fontitalic', defaults['fontitalic']).lower() == 'true'
    self.defaultFontParams = {
      'name' : self.params.get('fontName',fontName),
      'size' : int( self.params.get('fontSize',fontSize) ),
      'bold' : self.params.get('fontBold',fontBold),
      'italic' : self.params.get('fontItalic',fontItalic),
    }

  def output(self, fileObj):
    if self.outputFormat == 'png':
      self.surface.write_to_png(fileObj)
    elif self.outputFormat == 'pdf':
      self.surface.finish()
      pdfData = self.surfaceData.getvalue()
      self.surfaceData.close()
      fileObj.write(pdfData)
    else:
      if hasattr(self, 'startTime'):
        hasData = True
        metaData = {
          'x': {
            'start': self.startTime,
            'end': self.endTime
          },
          'options': {
            'lineWidth': self.lineWidth
          },
          'font': self.defaultFontParams,
          'area': self.area,
          'series': []
        }

        if not self.secondYAxis:
          metaData['y'] = {
            'top': self.yTop,
            'bottom': self.yBottom,
            'step': self.yStep,
            'labels': self.yLabels,
            'labelValues': self.yLabelValues
          }

        for series in self.data:
          if 'stacked' not in series.options:
            metaData['series'].append({
              'name': series.name,
              'start': series.start,
              'end': series.end,
              'step': series.step,
              'valuesPerPoint': series.valuesPerPoint,
              'color': series.color,
              'data': series,
              'options': series.options
            })
      else:
        hasData = False
        metaData = { }

      self.surface.finish()
      svgData = self.surfaceData.getvalue().decode('utf-8')
      self.surfaceData.close()

      svgData = svgData.replace('pt"', 'px"', 2)  # we expect height/width in pixels, not points
      svgData = svgData.replace('</svg>\n', '', 1)
      svgData = svgData.replace('</defs>\n<g', '</defs>\n<g class="graphite"', 1)

      if hasData:
        # We encode headers using special paths with d^="M -88 -88"
        # Find these, and turn them into <g> wrappers instead
        def onHeaderPath(match):
          name = ''
          for char in re.findall(r'L -(\d+) -\d+', match.group(1)):
            name += chr(int(char))
          return '</g><g data-header="true" class="%s">' % name
        (svgData, subsMade) = re.subn(r'<path.+?d="M -88 -88 (.+?)"/>', onHeaderPath, svgData)

        # Replace the first </g><g> with <g>, and close out the last </g> at the end
        svgData = svgData.replace('</g><g data-header', '<g data-header', 1)
        if subsMade > 0:
          svgData += '</g>'
        svgData = svgData.replace(' data-header="true"', '')

      fileObj.write(svgData.encode('utf-8'))
      fileObj.write(("""<script>
  <![CDATA[
    metadata = %s
  ]]>
</script>
</svg>""" % json.dumps(metaData)).encode('utf-8'))


class LineGraph(Graph):
  customizable = Graph.customizable + \
                 ('title','vtitle','lineMode','lineWidth','hideLegend',
                  'hideAxes','minXStep','hideGrid','majorGridLineColor',
                  'minorGridLineColor','thickness','min','max',
                  'graphOnly','yMin','yMax','yLimit','yStep','areaMode',
                  'areaAlpha','drawNullAsZero','tz', 'yAxisSide','pieMode',
                  'yUnitSystem', 'logBase','yMinLeft','yMinRight','yMaxLeft',
                  'yMaxRight', 'yLimitLeft', 'yLimitRight', 'yStepLeft',
                  'yStepRight', 'rightWidth', 'rightColor', 'rightDashed',
                  'leftWidth', 'leftColor', 'leftDashed', 'xFormat', 'minorY',
                  'hideYAxis', 'uniqueLegend', 'vtitleRight', 'yDivisors',
                  'connectedLimit', 'hideXAxis', 'hideNullFromLegend')
  validLineModes = ('staircase','slope','connected')
  validAreaModes = ('none','first','all','stacked')
  validPieModes = ('maximum', 'minimum', 'average')

  def drawGraph(self,**params):
    # Make sure we've got datapoints to draw
    if self.data:
      startTime = min([series.start for series in self.data])
      endTime = max([series.end for series in self.data])
      timeRange = endTime - startTime
    else:
      timeRange = None

    if not timeRange:
      x = self.width / 2
      y = self.height / 2
      self.setColor('red')
      self.setFont(size=math.log(self.width * self.height) )
      self.drawText("No Data", x, y, align='center')
      return

    # Determine if we're doing a 2 y-axis graph.
    for series in self.data:
      if 'secondYAxis' in series.options:
        self.dataRight.append(series)
      else:
        self.dataLeft.append(series)
    if len(self.dataRight) > 0:
      self.secondYAxis = True

    # API compatibility hacks
    if params.get('graphOnly',False):
      params['hideLegend'] = True
      params['hideGrid'] = True
      params['hideAxes'] = True
      params['hideXAxis'] = False
      params['hideYAxis'] = False
      params['yAxisSide'] = 'left'
      params['title'] = ''
      params['vtitle'] = ''
      params['margin'] = 0
      params['tz'] = ''
      self.margin = 0
      self.area['xmin'] = 0
      self.area['xmax'] = self.width
      self.area['ymin'] = 0
      self.area['ymax'] = self.height
    if 'yMin' not in params and 'min' in params:
      params['yMin'] = params['min']
    if 'yMax' not in params and 'max' in params:
      params['yMax'] = params['max']
    if 'lineWidth' not in params and 'thickness' in params:
      params['lineWidth'] = params['thickness']
    if 'yAxisSide' not in params:
      params['yAxisSide'] = 'left'
    if 'yUnitSystem' not in params:
      params['yUnitSystem'] = 'si'
    else:
      params['yUnitSystem'] = six.text_type(params['yUnitSystem']).lower()
      if params['yUnitSystem'] not in UnitSystems:
        params['yUnitSystem'] = 'si'

    self.params = params

    # Don't do any of the special right y-axis stuff if we're drawing 2 y-axes.
    if self.secondYAxis:
      params['yAxisSide'] = 'left'

    # When Y Axis is labeled on the right, we subtract x-axis positions from the max,
    # instead of adding to the minimum
    if self.params.get('yAxisSide') == 'right':
      self.margin = self.width

    # Now to setup our LineGraph specific options
    self.lineWidth = float( params.get('lineWidth', 1.2) )
    self.lineMode = params.get('lineMode','slope').lower()
    self.connectedLimit = params.get("connectedLimit", INFINITY)
    assert self.lineMode in self.validLineModes, "Invalid line mode!"
    self.areaMode = params.get('areaMode','none').lower()
    assert self.areaMode in self.validAreaModes, "Invalid area mode!"
    self.pieMode = params.get('pieMode', 'maximum').lower()
    assert self.pieMode in self.validPieModes, "Invalid pie mode!"

    # Line mode slope does not work (or even make sense) for series that have
    # only one datapoint. So if any series have one datapoint we force staircase mode.
    if self.lineMode == 'slope':
      for series in self.data:
        if len(series) == 1:
          self.lineMode = 'staircase'
          break

    if self.secondYAxis:
      for series in self.data:
        if 'secondYAxis' in series.options:
          if 'rightWidth' in params:
            series.options['lineWidth'] = params['rightWidth']
          if 'rightDashed' in params:
            series.options['dashed'] = params['rightDashed']
          if 'rightColor' in params:
            series.color = params['rightColor']
        else:
          if 'leftWidth' in params:
            series.options['lineWidth'] = params['leftWidth']
          if 'leftDashed' in params:
            series.options['dashed'] = params['leftDashed']
          if 'leftColor' in params:
            series.color = params['leftColor']

    for series in self.data:
      if not hasattr(series, 'color'):
        series.color = next(self.colors)

    titleSize = self.defaultFontParams['size'] + math.floor( math.log(self.defaultFontParams['size']) )
    self.setFont( size=titleSize )
    self.setColor( self.foregroundColor )

    if params.get('title'):
      self.drawTitle( six.text_type( unquote_plus(params['title']) ) )
    if params.get('vtitle'):
      self.drawVTitle( six.text_type( unquote_plus(params['vtitle']) ) )
    if self.secondYAxis and params.get('vtitleRight'):
      self.drawVTitle( six.text_type( unquote_plus(params['vtitleRight']) ), rightAlign=True )
    self.setFont()

    if not params.get('hideLegend', len(self.data) > settings.LEGEND_MAX_ITEMS):
      elements = []
      for series in self.data:
        if series.name:
          if not (params.get('hideNullFromLegend', False) and all(v is None for v in list(series))):
            elements.append((unquote_plus(series.name),series.color,series.options.get('secondYAxis')))
      if len(elements) > 0:
        self.drawLegend(elements, params.get('uniqueLegend', False))

    # Setup axes, labels, and grid
    # First we adjust the drawing area size to fit X-axis labels
    if not self.params.get('hideAxes',False) and not self.params.get('hideXAxis', False):
      self.area['ymax'] -= self.getExtents()['maxAscent'] * 2

    self.startTime = min([series.start for series in self.data])
    if self.lineMode == 'staircase':
      self.endTime = max([series.end for series in self.data])
    else:
      self.endTime = max([(series.end - series.step) for series in self.data])
    self.timeRange = self.endTime - self.startTime

    # Now we consolidate our data points to fit in the currently estimated drawing area
    self.consolidateDataPoints()

    self.encodeHeader('axes')

    # Now its time to fully configure the Y-axis and determine the space required for Y-axis labels
    # Since we'll probably have to squeeze the drawing area to fit the Y labels, we may need to
    # reconsolidate our data points, which in turn means re-scaling the Y axis, this process will
    # repeat until we have accurate Y labels and enough space to fit our data points
    currentXMin = self.area['xmin']
    currentXMax = self.area['xmax']

    if self.secondYAxis:
      self.setupTwoYAxes()
    else:
      self.setupYAxis()
    while currentXMin != self.area['xmin'] or currentXMax != self.area['xmax']: #see if the Y-labels require more space
      self.consolidateDataPoints() #this can cause the Y values to change
      currentXMin = self.area['xmin'] #so let's keep track of the previous Y-label space requirements
      currentXMax = self.area['xmax']
      if self.secondYAxis: #and recalculate their new requirements
        self.setupTwoYAxes()
      else:
        self.setupYAxis()

    # Now that our Y-axis is finalized, let's determine our X labels (this won't affect the drawing area)
    self.setupXAxis()

    if not self.params.get('hideAxes',False):
      self.drawLabels()
      if not self.params.get('hideGrid',False): #hideAxes implies hideGrid
        self.encodeHeader('grid')
        self.drawGridLines()

    # Finally, draw the graph lines
    self.encodeHeader('lines')
    self.drawLines()

  def drawVTitle(self, text, rightAlign=False):
    lineHeight = self.getExtents()['maxHeight']

    if rightAlign:
      self.encodeHeader('vtitleRight')
      x = self.area['xmax'] - lineHeight
      y = self.height / 2
      for line in text.split('\n'):
        self.drawText(line, x, y, align='center', valign='baseline', rotate=90)
        x -= lineHeight
      self.area['xmax'] = x - self.margin - lineHeight
    else:
      self.encodeHeader('vtitle')
      x = self.area['xmin'] + lineHeight
      y = self.height / 2
      for line in text.split('\n'):
        self.drawText(line, x, y, align='center', valign='baseline', rotate=270)
        x += lineHeight
      self.area['xmin'] = x + self.margin + lineHeight

  def getYCoord(self, value, side=None):
    if "left" == side:
      yLabelValues = self.yLabelValuesL
      yTop = self.yTopL
      yBottom = self.yBottomL
    elif "right" == side:
      yLabelValues = self.yLabelValuesR
      yTop = self.yTopR
      yBottom = self.yBottomR
    else:
      yLabelValues = self.yLabelValues
      yTop = self.yTop
      yBottom = self.yBottom

    try:
      highestValue = max(yLabelValues)
      lowestValue = min(yLabelValues)
    except ValueError:
      highestValue = yTop
      lowestValue = yBottom

    pixelRange = self.area['ymax'] - self.area['ymin']

    relativeValue = value - lowestValue
    valueRange = highestValue - lowestValue

    if self.logBase:
        if value <= 0:
            return None
        relativeValue = math.log(value, self.logBase) - math.log(lowestValue, self.logBase)
        valueRange = math.log(highestValue, self.logBase) - math.log(lowestValue, self.logBase)

    pixelToValueRatio = pixelRange / valueRange
    valueInPixels = pixelToValueRatio * relativeValue
    return self.area['ymax'] - valueInPixels

  def drawLines(self, width=None, dash=None, linecap='butt', linejoin='miter'):
    if not width: width = self.lineWidth
    self.ctx.set_line_width(width)
    originalWidth = width
    width = float(int(width) % 2) / 2
    if dash:
      self.ctx.set_dash(dash,1)
    else:
      self.ctx.set_dash([],0)
    self.ctx.set_line_cap({
      'butt' : cairo.LINE_CAP_BUTT,
      'round' : cairo.LINE_CAP_ROUND,
      'square' : cairo.LINE_CAP_SQUARE,
    }[linecap])
    self.ctx.set_line_join({
      'miter' : cairo.LINE_JOIN_MITER,
      'round' : cairo.LINE_JOIN_ROUND,
      'bevel' : cairo.LINE_JOIN_BEVEL,
    }[linejoin])

    # check whether there is an stacked metric
    singleStacked = False
    for series in self.data:
      if 'stacked' in series.options:
        singleStacked = True
    if singleStacked:
      self.data = sort_stacked(self.data)

    # stack the values
    if self.areaMode == 'stacked' and not self.secondYAxis: #TODO Allow stacked area mode with secondYAxis
      total = []
      for series in self.data:
        if 'drawAsInfinite' in series.options:
          continue

        series.options['stacked'] = True
        for i in range(len(series)):
          if len(total) <= i: total.append(0)

          if series[i] is not None:
            original = series[i]
            series[i] += total[i]
            total[i] += original
    elif self.areaMode == 'first':
      self.data[0].options['stacked'] = True
    elif self.areaMode == 'all':
      for series in self.data:
        if 'drawAsInfinite' not in series.options:
          series.options['stacked'] = True

    # apply alpha channel and create separate stroke series
    if self.params.get('areaAlpha'):
      try:
        alpha = float(self.params['areaAlpha'])
      except ValueError:
        alpha = 0.5
        pass

      strokeSeries = []
      for series in self.data:
        if 'stacked' in series.options:
          series.options['alpha'] = alpha

          newSeries = TimeSeries(series.name, series.start, series.end, series.step*series.valuesPerPoint, [x for x in series])
          newSeries.xStep = series.xStep
          newSeries.color = series.color
          if 'secondYAxis' in series.options:
            newSeries.options['secondYAxis'] = True
          strokeSeries.append(newSeries)
      self.data += strokeSeries

    # setup the clip region
    self.ctx.set_line_width(1.0)
    self.ctx.rectangle(self.area['xmin'], self.area['ymin'], self.area['xmax'] - self.area['xmin'], self.area['ymax'] - self.area['ymin'])
    self.ctx.clip()
    self.ctx.set_line_width(originalWidth)

    # save clip to restore once stacked areas are drawn
    self.ctx.save()
    clipRestored = False

    for series in self.data:

      if 'stacked' not in series.options:
        # stacked areas are always drawn first. if this series is not stacked, we finished stacking.
        # reset the clip region so lines can show up on top of the stacked areas.
        if not clipRestored:
          clipRestored = True
          self.ctx.restore()

      if 'lineWidth' in series.options:
        self.ctx.set_line_width(series.options['lineWidth'])

      if 'dashed' in series.options:
        self.ctx.set_dash([ series.options['dashed'] ], 1)
      else:
        self.ctx.set_dash([], 0)

      # Shift the beginning of drawing area to the start of the series if the
      # graph itself has a larger range
      missingPoints = (series.start - self.startTime) / series.step
      startShift = series.xStep * (missingPoints / series.valuesPerPoint)
      x = float(self.area['xmin']) + startShift + (self.lineWidth / 2.0)
      y = float(self.area['ymin'])

      startX = x

      if series.options.get('invisible'):
        self.setColor( series.color, 0, True )
      else:
        self.setColor( series.color, series.options.get('alpha') or 1.0 )

      # The number of preceding datapoints that had a None value.
      consecutiveNones = 0

      for index, value in enumerate(series):
        if value != value: # convert NaN to None
          value = None

        if value is None and self.params.get('drawNullAsZero'):
          value = 0.0

        if value is None:
          if consecutiveNones == 0:
            self.ctx.line_to(x, y)
            if 'stacked' in series.options: #Close off and fill area before unknown interval
              if self.secondYAxis:
                if 'secondYAxis' in series.options:
                  self.fillAreaAndClip(x, y, startX, self.getYCoord(0, "right"))
                else:
                  self.fillAreaAndClip(x, y, startX, self.getYCoord(0, "left"))
              else:
                self.fillAreaAndClip(x, y, startX, self.getYCoord(0))

          x += series.xStep
          consecutiveNones += 1

        else:
          if self.secondYAxis:
            if 'secondYAxis' in series.options:
              y = self.getYCoord(value, "right")
            else:
              y = self.getYCoord(value, "left")
          else:
            y = self.getYCoord(value)

          if y is None:
            value = None
          elif y < 0:
              y = 0

          if 'drawAsInfinite' in series.options and value > 0:
            self.ctx.move_to(x, self.area['ymax'])
            self.ctx.line_to(x, self.area['ymin'])
            self.ctx.stroke()
            x += series.xStep
            continue

          if consecutiveNones > 0:
            startX = x

          if self.lineMode == 'staircase':
            if consecutiveNones > 0:
              self.ctx.move_to(x, y)
            else:
              self.ctx.line_to(x, y)

            x += series.xStep
            self.ctx.line_to(x, y)

          elif self.lineMode == 'slope':
            if consecutiveNones > 0:
              self.ctx.move_to(x, y)

            self.ctx.line_to(x, y)
            x += series.xStep

          elif self.lineMode == 'connected':
            # If if the gap is larger than the connectedLimit or if this is the
            # first non-None datapoint in the series, start drawing from that datapoint.
            if consecutiveNones > self.connectedLimit or consecutiveNones == index:
               self.ctx.move_to(x, y)

            self.ctx.line_to(x, y)
            x += series.xStep

          consecutiveNones = 0

      if 'stacked' in series.options:
        if self.lineMode == 'staircase':
          xPos = x
        else:
          xPos = x-series.xStep
        if self.secondYAxis:
          if 'secondYAxis' in series.options:
            areaYFrom = self.getYCoord(0, "right")
          else:
            areaYFrom = self.getYCoord(0, "left")
        else:
          areaYFrom = self.getYCoord(0)

        self.fillAreaAndClip(xPos, y, startX, areaYFrom)

      else:
        self.ctx.stroke()

      self.ctx.set_line_width(originalWidth) # return to the original line width
      if 'dash' in series.options: # if we changed the dash setting before, change it back now
        if dash:
          self.ctx.set_dash(dash,1)
        else:
          self.ctx.set_dash([],0)

  def fillAreaAndClip(self, x, y, startX=None, areaYFrom=None):
    startX = (startX or self.area['xmin'])
    areaYFrom = (areaYFrom or self.area['ymax'])
    pattern = self.ctx.copy_path()

    # fill
    self.ctx.line_to(x, areaYFrom)                  # bottom endX
    self.ctx.line_to(startX, areaYFrom)             # bottom startX
    self.ctx.close_path()
    self.ctx.fill()

    # clip above y axis
    self.ctx.append_path(pattern)
    self.ctx.line_to(x, areaYFrom)                  # yZero endX
    self.ctx.line_to(self.area['xmax'], areaYFrom)  # yZero right
    self.ctx.line_to(self.area['xmax'], self.area['ymin'])  # top right
    self.ctx.line_to(self.area['xmin'], self.area['ymin'])  # top left
    self.ctx.line_to(self.area['xmin'], areaYFrom)  # yZero left
    self.ctx.line_to(startX, areaYFrom)             # yZero startX

    # clip below y axis
    self.ctx.line_to(x, areaYFrom)                  # yZero endX
    self.ctx.line_to(self.area['xmax'], areaYFrom)  # yZero right
    self.ctx.line_to(self.area['xmax'], self.area['ymax'])  # bottom right
    self.ctx.line_to(self.area['xmin'], self.area['ymax'])  # bottom left
    self.ctx.line_to(self.area['xmin'], areaYFrom)  # yZero left
    self.ctx.line_to(startX, areaYFrom)             # yZero startX
    self.ctx.close_path()
    self.ctx.clip()

  def consolidateDataPoints(self):
    numberOfPixels = self.graphWidth = self.area['xmax'] - self.area['xmin'] - (self.lineWidth + 1)
    for series in self.data:
      numberOfDataPoints = self.timeRange/series.step
      minXStep = float( self.params.get('minXStep',1.0) )
      divisor = self.timeRange / series.step
      bestXStep = numberOfPixels / divisor
      if bestXStep < minXStep:
        drawableDataPoints = int( numberOfPixels / minXStep )
        pointsPerPixel = math.ceil( float(numberOfDataPoints) / float(drawableDataPoints) )
        series.consolidate(pointsPerPixel)
        series.xStep = (numberOfPixels * pointsPerPixel) / numberOfDataPoints
      else:
        series.xStep = bestXStep

  def _adjustLimits(self, minValue, maxValue, minName, maxName, limitName):
    if maxName in self.params and self.params[maxName] != 'max':
      maxValue = self.params[maxName]

    if limitName in self.params and self.params[limitName] < maxValue:
      maxValue = self.params[limitName]

    if minName in self.params:
      minValue = self.params[minName]

    if maxValue <= minValue:
      maxValue = minValue + 1

    return (minValue, maxValue)

  def setupYAxis(self):
    (yMinValue, yMaxValue) = dataLimits(self.data,
                                        drawNullAsZero=self.params.get('drawNullAsZero'),
                                        stacked=(self.areaMode == 'stacked'))

    if self.logBase:
      yTics = _LogAxisTics(yMinValue, yMaxValue,
                           unitSystem=self.params.get('yUnitSystem'), base=self.logBase)
    else:
      yTics = _LinearAxisTics(yMinValue, yMaxValue,
                           unitSystem=self.params.get('yUnitSystem'))

    yTics.applySettings(axisMin=self.params.get('yMin'),
                        axisMax=self.params.get('yMax'),
                        axisLimit=self.params.get('yLimit'))

    if 'yStep' in self.params:
      yTics.setStep(self.params['yStep'])
    else:
      yDivisors = str(self.params.get('yDivisors', '4,5,6'))
      yDivisors = [int(d) for d in yDivisors.split(',')]
      binary = self.params.get('yUnitSystem') == 'binary'
      yTics.chooseStep(divisors=yDivisors, binary=binary)

    yTics.chooseLimits()

    # Copy the values we need back out of the yTics object:
    self.yStep = yTics.step
    self.yBottom = yTics.bottom
    self.yTop = yTics.top
    self.ySpan = yTics.span

    if not self.params.get('hideAxes', False):
      #Create and measure the Y-labels

      self.yLabelValues = yTics.getLabelValues()
      self.yLabels = [yTics.makeLabel(value) for value in self.yLabelValues]
      self.yLabelWidth = max([self.getExtents(label)['width'] for label in self.yLabels])

      if not self.params.get('hideYAxis'):
        if self.params.get('yAxisSide') == 'left':
          # Scoot the graph over to the left just enough to fit the y-labels:
          xMin = self.margin + (self.yLabelWidth * 1.02)
          if self.area['xmin'] < xMin:
            self.area['xmin'] = xMin
        else:
          # Scoot the graph over to the right just enough to fit the y-labels:
          xMin = 0
          xMax = self.margin - (self.yLabelWidth * 1.02)
          if self.area['xmax'] >= xMax:
            self.area['xmax'] = xMax
    else:
      self.yLabelValues = []
      self.yLabels = []
      self.yLabelWidth = 0.0

  def setupTwoYAxes(self):
    drawNullAsZero = self.params.get('drawNullAsZero')
    stacked = (self.areaMode == 'stacked')

    (yMinValueL, yMaxValueL) = dataLimits(self.dataLeft, drawNullAsZero, stacked)
    (yMinValueR, yMaxValueR) = dataLimits(self.dataRight, drawNullAsZero, stacked)

    # TODO: Allow separate bases for L & R Axes.
    if self.logBase:
      yTicsL = _LogAxisTics(yMinValueL, yMaxValueL,
                            unitSystem=self.params.get('yUnitSystem'), base=self.logBase)
      yTicsR = _LogAxisTics(yMinValueR, yMaxValueR,
                            unitSystem=self.params.get('yUnitSystem'), base=self.logBase)
    else:
      yTicsL = _LinearAxisTics(yMinValueL, yMaxValueL,
                               unitSystem=self.params.get('yUnitSystem'))
      yTicsR = _LinearAxisTics(yMinValueR, yMaxValueR,
                               unitSystem=self.params.get('yUnitSystem'))

    yTicsL.applySettings(axisMin=self.params.get('yMinLeft'),
                         axisMax=self.params.get('yMaxLeft'),
                         axisLimit=self.params.get('yLimitLeft'))
    yTicsR.applySettings(axisMin=self.params.get('yMinRight'),
                         axisMax=self.params.get('yMaxRight'),
                         axisLimit=self.params.get('yLimitRight'))

    yDivisors = str(self.params.get('yDivisors', '4,5,6'))
    yDivisors = [int(d) for d in yDivisors.split(',')]
    binary = self.params.get('yUnitSystem') == 'binary'

    if 'yStepLeft' in self.params:
      yTicsL.setStep(self.params['yStepLeft'])
    else:
      yTicsL.chooseStep(divisors=yDivisors, binary=binary)

    if 'yStepRight' in self.params:
      yTicsR.setStep(self.params['yStepRight'])
    else:
      yTicsR.chooseStep(divisors=yDivisors, binary=binary)

    yTicsL.chooseLimits()
    yTicsR.chooseLimits()

    # Copy the values we need back out of the yTics objects:
    self.yStepL = yTicsL.step
    self.yBottomL = yTicsL.bottom
    self.yTopL = yTicsL.top
    self.ySpanL = yTicsL.span

    self.yStepR = yTicsR.step
    self.yBottomR = yTicsR.bottom
    self.yTopR = yTicsR.top
    self.ySpanR = yTicsR.span

    #Create and measure the Y-labels
    self.yLabelValuesL = yTicsL.getLabelValues()
    self.yLabelValuesR = yTicsR.getLabelValues()

    self.yLabelsL = [yTicsL.makeLabel(value) for value in self.yLabelValuesL]
    self.yLabelsR = [yTicsR.makeLabel(value) for value in self.yLabelValuesR]

    self.yLabelWidthL = max([self.getExtents(label)['width'] for label in self.yLabelsL])
    self.yLabelWidthR = max([self.getExtents(label)['width'] for label in self.yLabelsR])
    #scoot the graph over to the left just enough to fit the y-labels

    #xMin = self.margin + self.margin + (self.yLabelWidthL * 1.02)
    xMin = self.margin + (self.yLabelWidthL * 1.02)
    if self.area['xmin'] < xMin:
      self.area['xmin'] = xMin
    #scoot the graph over to the right just enough to fit the y-labels
    xMax = self.width - (self.yLabelWidthR * 1.02)
    if self.area['xmax'] >= xMax:
      self.area['xmax'] = xMax

  def setupXAxis(self):
    if self.userTimeZone:
      tzinfo = pytz.timezone(self.userTimeZone)
    else:
      tzinfo = pytz.timezone(settings.TIME_ZONE)

    self.start_dt = datetime.fromtimestamp(self.startTime, tzinfo)
    self.end_dt = datetime.fromtimestamp(self.endTime, tzinfo)

    secondsPerPixel = float(self.timeRange) / float(self.graphWidth)
    self.xScaleFactor = float(self.graphWidth) / float(self.timeRange) #pixels per second

    potential = [c for c in xAxisConfigs if c['seconds'] <= secondsPerPixel and c.get('maxInterval', self.timeRange + 1) >= self.timeRange]
    if potential:
      self.xConf = potential[-1]
    else:
      self.xConf = xAxisConfigs[-1]

    self.xLabelStep = self.xConf['labelUnit'] * self.xConf['labelStep']
    self.xMinorGridStep = self.xConf['minorGridUnit'] * self.xConf['minorGridStep']
    self.xMajorGridStep = self.xConf['majorGridUnit'] * self.xConf['majorGridStep']

  def drawLabels(self):
    # Draw the Y-labels
    if not self.params.get('hideYAxis'):
      if not self.secondYAxis:
        for value,label in zip(self.yLabelValues,self.yLabels):
          if self.params.get('yAxisSide') == 'left':
            x = self.area['xmin'] - (self.yLabelWidth * 0.02)
          else:
            x = self.area['xmax'] + (self.yLabelWidth * 0.02) #Inverted for right side Y Axis

          y = self.getYCoord(value)
          if y is None:
              value = None
          elif y < 0:
              y = 0

          if self.params.get('yAxisSide') == 'left':
            self.drawText(label, x, y, align='right', valign='middle')
          else:
            self.drawText(label, x, y, align='left', valign='middle') #Inverted for right side Y Axis
      else: # Draws a right side and a Left side axis
        for valueL,labelL in zip(self.yLabelValuesL,self.yLabelsL):
          xL = self.area['xmin'] - (self.yLabelWidthL * 0.02)
          yL = self.getYCoord(valueL, "left")
          if yL is None:
            value = None
          elif yL < 0:
            yL = 0
          self.drawText(labelL, xL, yL, align='right', valign='middle')

          ### Right Side
        for valueR,labelR in zip(self.yLabelValuesR,self.yLabelsR):
          xR = self.area['xmax'] + (self.yLabelWidthR * 0.02) + 3 #Inverted for right side Y Axis
          yR = self.getYCoord(valueR, "right")
          if yR is None:
            valueR = None
          elif yR < 0:
            yR = 0
          self.drawText(labelR, xR, yR, align='left', valign='middle') #Inverted for right side Y Axis

    if not self.params.get('hideXAxis'):
      (dt, x_label_delta) = find_x_times(self.start_dt, self.xConf['labelUnit'], self.xConf['labelStep'])

      # Draw the X-labels
      xFormat = self.params.get('xFormat', self.xConf['format'])
      while dt < self.end_dt:
        label = dt.strftime(xFormat)
        x = self.area['xmin'] + (toSeconds(dt - self.start_dt) * self.xScaleFactor)
        y = self.area['ymax'] + self.getExtents()['maxAscent']
        self.drawText(label, x, y, align='center', valign='top')
        dt += x_label_delta

  def drawGridLines(self):
    # Not sure how to handle this for 2 y-axes
    # Just using the left side info for the grid.

    # Horizontal grid lines
    leftSide = self.area['xmin']
    rightSide = self.area['xmax']
    labels = []
    if self.secondYAxis:
      labels = self.yLabelValuesL
    else:
      labels = self.yLabelValues

    for i, value in enumerate(labels):
      self.ctx.set_line_width(0.4)
      self.setColor( self.params.get('majorGridLineColor',self.defaultMajorGridLineColor) )

      if self.secondYAxis:
        y = self.getYCoord(value,"left")
      else:
        y = self.getYCoord(value)

      if y is None or y < 0:
          continue
      self.ctx.move_to(leftSide, y)
      self.ctx.line_to(rightSide, y)
      self.ctx.stroke()

      # draw minor gridlines if this isn't the last label
      if self.minorY >= 1 and i < (len(labels) - 1):
        # in case graphite supports inverted Y axis now or someday
        (valueLower, valueUpper) = sorted((value, labels[i+1]))

        # each minor gridline is 1/minorY apart from the nearby gridlines.
        # we calculate that distance, for adding to the value in the loop.
        distance = ((valueUpper - valueLower) / float(1 + self.minorY))

        # starting from the initial valueLower, we add the minor distance
        # for each minor gridline that we wish to draw, and then draw it.
        for minor in range(self.minorY):
          self.ctx.set_line_width(0.3)
          self.setColor( self.params.get('minorGridLineColor',self.defaultMinorGridLineColor) )

          # the current minor gridline value is halfway between the current and next major gridline values
          value = (valueLower + ((1+minor) * distance))

          if self.logBase:
            yTopFactor = self.logBase * self.logBase
          else:
            yTopFactor = 1

          if self.secondYAxis:
            if value >= (yTopFactor * self.yTopL):
              continue
          else:
            if value >= (yTopFactor * self.yTop):
              continue

          if self.secondYAxis:
            y = self.getYCoord(value,"left")
          else:
            y = self.getYCoord(value)
          if y is None or y < 0:
              continue

          self.ctx.move_to(leftSide, y)
          self.ctx.line_to(rightSide, y)
          self.ctx.stroke()

    # Vertical grid lines
    top = self.area['ymin']
    bottom = self.area['ymax']

    # First we do the minor grid lines (majors will paint over them)
    self.ctx.set_line_width(0.25)
    self.setColor( self.params.get('minorGridLineColor',self.defaultMinorGridLineColor) )
    (dt, x_minor_delta) = find_x_times(self.start_dt, self.xConf['minorGridUnit'], self.xConf['minorGridStep'])

    while dt < self.end_dt:
      x = self.area['xmin'] + (toSeconds(dt - self.start_dt) * self.xScaleFactor)

      if x < self.area['xmax']:
        self.ctx.move_to(x, bottom)
        self.ctx.line_to(x, top)
        self.ctx.stroke()

      dt += x_minor_delta

    # Now we do the major grid lines
    self.ctx.set_line_width(0.33)
    self.setColor( self.params.get('majorGridLineColor',self.defaultMajorGridLineColor) )
    (dt, x_major_delta) = find_x_times(self.start_dt, self.xConf['majorGridUnit'], self.xConf['majorGridStep'])

    while dt < self.end_dt:
      x = self.area['xmin'] + (toSeconds(dt - self.start_dt) * self.xScaleFactor)

      if x < self.area['xmax']:
        self.ctx.move_to(x, bottom)
        self.ctx.line_to(x, top)
        self.ctx.stroke()

      dt += x_major_delta

    # Draw side borders for our graph area
    self.ctx.set_line_width(0.5)
    self.ctx.move_to(self.area['xmax'], bottom)
    self.ctx.line_to(self.area['xmax'], top)
    self.ctx.move_to(self.area['xmin'], bottom)
    self.ctx.line_to(self.area['xmin'], top)
    self.ctx.stroke()


class PieGraph(Graph):
  customizable = Graph.customizable + \
                 ('title','valueLabels','valueLabelsMin','hideLegend','pieLabels','areaAlpha','valueLabelsColor')
  validValueLabels = ('none','number','percent')

  def drawGraph(self,**params):
    self.pieLabels = params.get('pieLabels', 'horizontal')
    self.total = sum( [t[1] for t in self.data] )

    if not self.data:
      x = self.width / 2
      y = self.height / 2
      self.setColor('red')
      self.setFont(size=math.log(self.width * self.height) )
      self.drawText("No Data", x, y, align='center')
      return

    if self.params.get('areaAlpha'):
      try:
        self.alpha = float(self.params['areaAlpha'])
      except ValueError:
        self.alpha = 1.0
        pass
    else:
      self.alpha = 1.0

    self.slices = []
    for name,value in self.data:
      self.slices.append({
        'name' : name,
        'value' : value,
        'percent' : value / self.total,
        'color' : next(self.colors),
        'alpha' : self.alpha,
      })

    titleSize = self.defaultFontParams['size'] + math.floor( math.log(self.defaultFontParams['size']) )
    self.setFont( size=titleSize )
    self.setColor( self.foregroundColor )
    if params.get('title'):
      self.drawTitle( unquote_plus(params['title']) )
    self.setFont()

    if not params.get('hideLegend',False):
      elements = [ (slice['name'],slice['color'],None) for slice in self.slices ]
      if len(elements) > 0:
        self.drawLegend(elements)

    self.drawSlices()

    if params.get('valueLabelsColor'):
      self.valueLabelsColor = params.get('valueLabelsColor')
    else:
      self.valueLabelsColor = 'black'

    self.valueLabelsMin = float( params.get('valueLabelsMin',5) )
    self.valueLabels = params.get('valueLabels','percent')
    assert self.valueLabels in self.validValueLabels, \
        "valueLabels=%s must be one of %s" % (self.valueLabels,self.validValueLabels)
    if self.valueLabels != 'none':
      self.drawLabels()

  def drawSlices(self):
    theta = 3.0 * math.pi / 2.0
    halfX = (self.area['xmax'] - self.area['xmin']) / 2.0
    halfY = (self.area['ymax'] - self.area['ymin']) / 2.0
    self.x0 = x0 = self.area['xmin'] + halfX
    self.y0 = y0 = self.area['ymin'] + halfY
    self.radius = radius = min(halfX,halfY) * 0.95
    for slice in self.slices:
      self.setColor( slice['color'], slice['alpha'] )
      self.ctx.move_to(x0,y0)
      phi = theta + (2 * math.pi) * slice['percent']
      self.ctx.arc( x0, y0, radius, theta, phi )
      self.ctx.line_to(x0,y0)
      self.ctx.fill()
      slice['midAngle'] = (theta + phi) / 2.0
      slice['midAngle'] %= 2.0 * math.pi
      theta = phi

  def drawLabels(self):
    self.setFont()
    self.setColor( self.valueLabelsColor )
    for slice in self.slices:
      if self.valueLabels == 'percent':
        if (slice['percent'] * 100.0) < self.valueLabelsMin: continue
        label = "%%%.2f" % (slice['percent'] * 100.0)
      elif self.valueLabels == 'number':
        if slice['value'] < self.valueLabelsMin: continue
        if slice['value'] < 10 and slice['value'] != int(slice['value']):
          label = "%.2f" % slice['value']
        else:
          label = six.text_type(int(slice['value']))
      theta = slice['midAngle']
      x = self.x0 + (self.radius / 2.0 * math.cos(theta))
      y = self.y0 + (self.radius / 2.0 * math.sin(theta))

      if self.pieLabels == 'rotated':
        if theta > (math.pi / 2.0) and theta <= (3.0 * math.pi / 2.0):
          theta -= math.pi
        self.drawText( label, x, y, align='center', valign='middle', rotate=math.degrees(theta) )
      else:
        self.drawText( label, x, y, align='center', valign='middle')


GraphTypes = {
  'line' : LineGraph,
  'pie' : PieGraph,
}


# Convenience functions
def toSeconds(t):
  return (t.days * 86400) + t.seconds


def safeArgs(args):
  """Iterate over valid, finite values in an iterable.

  Skip any items that are None, NaN, or infinite.
  """

  return (arg for arg in args
          if arg is not None and not math.isnan(arg) and not math.isinf(arg))


def safeMin(args):
  args = list(safeArgs(args))
  if args:
    return min(args)


def safeMax(args):
  args = list(safeArgs(args))
  if args:
    return max(args)


def safeSum(values):
  return sum(safeArgs(values))


def any(args):
  for arg in args:
    if arg:
      return True
  return False


def dataLimits(data, drawNullAsZero=False, stacked=False):
  """Return the range of values in data as (yMinValue, yMaxValue).

  data is an array of TimeSeries objects.
  """

  missingValues = any(None in series for series in data)
  finiteData = [series for series in data if not series.options.get('drawAsInfinite')]

  yMinValue = safeMin(safeMin(series) for series in finiteData)

  if yMinValue is None:
    # This can only happen if there are no valid, non-infinite data.
    return (0.0, 1.0)

  if yMinValue > 0.0 and drawNullAsZero and missingValues:
    yMinValue = 0.0

  if stacked:
    length = safeMin(len(series) for series in finiteData)
    sumSeries = []

    for i in range(0, length):
      sumSeries.append( safeSum(series[i] for series in finiteData) )
    yMaxValue = safeMax( sumSeries )
  else:
    yMaxValue = safeMax(safeMax(series) for series in finiteData)

  if yMaxValue < 0.0 and drawNullAsZero and missingValues:
    yMaxValue = 0.0

  return (yMinValue, yMaxValue)


def sort_stacked(series_list):
  stacked = [s for s in series_list if 'stacked' in s.options]
  not_stacked = [s for s in series_list if 'stacked' not in s.options]
  return stacked + not_stacked


def format_units(v, step=None, system='si', units=None):
  """Format the given value in standardized units.

  ``system`` is either 'binary' or 'si'

  For more info, see:
    http://en.wikipedia.org/wiki/SI_prefix
    http://en.wikipedia.org/wiki/Binary_prefix
  """

  if v is None:
    return v, ''

  if step is None:
    condition = lambda size: abs(v) >= size
  else:
    condition = lambda size: abs(v) >= size and step >= size

  for prefix, size in UnitSystems[system]:
    if condition(size):
      v2 = v / size
      if (v2 - math.floor(v2)) < 0.00000000001 and v > 1:
        v2 = float(math.floor(v2))
      if units:
        prefix = "%s%s" % (prefix, units)
      return v2, prefix

  if (v - math.floor(v)) < 0.00000000001 and v > 1 :
    v = float(math.floor(v))
  if units:
    prefix = units
  else:
    prefix = ''
  return v, prefix


def find_x_times(start_dt, unit, step):
  if not isinstance(start_dt, datetime):
    raise ValueError("Invalid start_dt: %s" % start_dt)
  if not isinstance(step, int) or not step > 0:
    if not isinstance(step, float) or unit != DAY or not step > 0.0:
      raise ValueError("Invalid step value: %s" % step)

  if unit == SEC:
    dt = start_dt.replace(second=start_dt.second - (start_dt.second % step))
    x_delta = timedelta(seconds=step)

  elif unit == MIN:
    dt = start_dt.replace(second=0, minute=start_dt.minute - (start_dt.minute % step))
    x_delta = timedelta(minutes=step)

  elif unit == HOUR:
    dt = start_dt.replace(second=0, minute=0, hour=start_dt.hour - (start_dt.hour % step))
    x_delta = timedelta(hours=step)

  elif unit == DAY:
    dt = start_dt.replace(second=0, minute=0, hour=0)
    x_delta = timedelta(days=step)

  else:
    raise ValueError("Invalid unit: %s" % unit)

  while dt < start_dt:
    dt += x_delta

  return (dt, x_delta)



from traceback import format_exc

from django.conf import settings

from graphite.logger import log
from graphite.readers.utils import BaseReader


class RemoteReader(BaseReader):
    __slots__ = (
        'finder',
        'metric_path',
        'intervals',
        'bulk_query',
    )

    def __init__(self, finder, node_info, bulk_query=None):
        self.finder = finder
        self.metric_path = node_info.get('path') or node_info.get('metric_path')
        self.intervals = node_info.get('intervals', [])
        self.bulk_query = set(bulk_query) if bulk_query else (
            [self.metric_path] if self.metric_path else []
        )

    def __repr__(self):
        return '<RemoteReader[%x]: %s %s>' % (id(self), self.finder.host, ','.join(self.bulk_query))

    def get_intervals(self):
        return self.intervals

    def fetch(self, startTime, endTime, now=None, requestContext=None):
        for series in self.fetch_multi(startTime, endTime, now, requestContext):
            if series['name'] == self.metric_path:
                return (series['time_info'], series['values'])

    def fetch_multi(self, startTime, endTime, now=None, requestContext=None):
        if not self.bulk_query:
            return []

        query_params = [
            ('format', self.finder.params.get('format', 'pickle')),
            ('local', self.finder.params.get('local', '1')),
            ('noCache', '1'),
            ('from', int(startTime)),
            ('until', int(endTime))
        ]

        for target in self.bulk_query:
            query_params.append(('target', target))

        if now is not None:
            query_params.append(('now', int(now)))

        headers = requestContext.get('forwardHeaders') if requestContext else None

        retries = 1  # start counting at one to make log output and settings more readable
        while True:
            try:
                result = self.finder.request(
                    '/render/',
                    fields=query_params,
                    headers=headers,
                    timeout=settings.FETCH_TIMEOUT,
                )
                break
            except Exception:
                if retries >= settings.MAX_FETCH_RETRIES:
                    log.exception("Failed after %s attempts! Root cause:\n%s" %
                                  (settings.MAX_FETCH_RETRIES, format_exc()))
                    raise
                else:
                    log.exception("Got an exception when fetching data! Try: %i of %i. Root cause:\n%s" %
                                  (retries, settings.MAX_FETCH_RETRIES, format_exc()))
                retries += 1

        data = self.finder.deserialize(result)

        try:
            return [
                {
                    'pathExpression': series.get('pathExpression', series['name']),
                    'name': series['name'],
                    'time_info': (series['start'], series['end'], series['step']),
                    'values': series['values'],
                }
                for series in data
            ]
        except Exception as err:
            self.finder.fail()
            log.exception("RemoteReader[%s] Invalid render response from %s: %s" %
                          (self.finder.host, result.url_full, repr(err)))
            raise Exception("Invalid render response from %s: %s" % (result.url_full, repr(err)))



import time
import sys

import six.moves.queue

from threading import Lock
from multiprocessing.pool import ThreadPool

_init_lock = Lock()
_pools = {}


class Job(object):
    """A job to be executed by a pool

    The job accepts a function and arguments.

    When it is run, it will execute the function with the specified arguments.

    The return value of the function will be stored in the result property of the job.

    If the function raises an exception, it will be stored in the exception property of the job.
    """
    __slots__ = (
        'func', 'description',
        'args', 'kwargs', 'result',
        'exception', 'exception_info',
    )

    def __init__(self, func, description, *args, **kwargs):
        self.func = func
        self.args = args
        self.description = description
        self.kwargs = kwargs
        self.result = None
        self.exception = None

    def __str__(self):
        return self.description

    def run(self):
        try:
            self.result = self.func(*self.args, **self.kwargs)
        except Exception as e:
            self.exception_info = sys.exc_info()
            self.exception = e


def get_pool(name="default", thread_count=1):
    """Get (and initialize) a Thread pool.

    If thread_count is 0, then None is returned.

    If the thread pool had already been initialized, thread_count will
    be ignored.
    """
    if not thread_count:
        return None

    with _init_lock:
        pool = _pools.get(name, None)
        if pool is None:
            pool = ThreadPool(thread_count)
            _pools[name] = pool
    return pool


def stop_pools():
    with _init_lock:
        for name in list(_pools.keys()):
            pool = _pools.pop(name)
            pool.close()


def stop_pool(name="default"):
    with _init_lock:
        _pools[name].close()
        del _pools[name]


class PoolTimeoutError(Exception):
    pass


def pool_exec(pool, jobs, timeout):
    """Execute a list of jobs, yielding each one as it completes.

    If a pool is specified then the jobs will be executed asynchronously,
    otherwise they are executed in order.

    If not all jobs have been executed after the specified timeout a
    PoolTimeoutError will be raised. When operating synchronously the
    timeout is checked before each job is run.
    """
    start = time.time()
    deadline = start + timeout
    if pool:
        queue = six.moves.queue.Queue()

        def pool_executor(job):
            job.run()
            queue.put(job)

        for job in jobs:
            pool.apply_async(func=pool_executor, args=[job])

        done = 0
        total = len(jobs)

        while done < total:
            wait_time = max(0, deadline - time.time())
            try:
                job = queue.get(True, wait_time)
            except six.moves.queue.Empty:
                raise PoolTimeoutError("Timed out after %fs" % (time.time() - start))

            done += 1
            yield job
    else:
        for job in jobs:
            if time.time() > deadline:
                raise PoolTimeoutError("Timed out after %fs" % (time.time() - start))

            job.run()
            yield job



from graphite.logger import log
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:  # Django < 1.10
    MiddlewareMixin = object


class LogExceptionsMiddleware(MiddlewareMixin):
    def process_exception(self, request, exception):
        log.exception('Exception encountered in <{0} {1}>'.format(request.method, request.build_absolute_uri()))
        return None



import traceback
from django.http import HttpResponseServerError
from django.template import loader


def server_error(request, template_name='500.html'):
    template = loader.get_template(template_name)
    context = {'stacktrace' : traceback.format_exc()}
    return HttpResponseServerError(template.render(context))



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from __future__ import division

import collections
import re
import time
import types
from six import text_type

from django.conf import settings

from graphite.logger import log
from graphite.storage import STORE
from graphite.util import timebounds, logtime


try:
  from collections import UserDict
except ImportError:
  from UserDict import IterableUserDict as UserDict


class Tags(UserDict):
  def __setitem__(self, key, value):
    self.data[key] = str(value)


class TimeSeries(list):
  def __init__(self, name, start, end, step, values, consolidate='average', tags=None, xFilesFactor=None, pathExpression=None):
    list.__init__(self, values)
    self.name = name
    self.start = start
    self.end = end
    self.step = step
    self.consolidationFunc = consolidate
    self.valuesPerPoint = 1
    self.options = {}
    self.pathExpression = pathExpression or name
    self.xFilesFactor = xFilesFactor if xFilesFactor is not None else settings.DEFAULT_XFILES_FACTOR

    if tags:
      self.tags = tags
    else:
      self.tags = {'name': name}
      # parse for tags if a tagdb is configured and name doesn't look like a function-wrapped name
      try:
        if STORE.tagdb and not re.match('^[a-z]+[(].+[)]$', name, re.IGNORECASE):
          self.tags = STORE.tagdb.parse(name).tags
      except Exception as err:
        # tags couldn't be parsed, just use "name" tag
        log.debug("Couldn't parse tags for %s: %s" % (name, err))

  def __eq__(self, other):
    if not isinstance(other, TimeSeries):
      return False

    if hasattr(self, 'color'):
      if not hasattr(other, 'color') or (self.color != other.color):
        return False
    elif hasattr(other, 'color'):
      return False

    return ((self.name, self.start, self.end, self.step, self.consolidationFunc, self.valuesPerPoint, self.options, self.xFilesFactor) ==
      (other.name, other.start, other.end, other.step, other.consolidationFunc, other.valuesPerPoint, other.options, other.xFilesFactor)) and list.__eq__(self, other)

  def __iter__(self):
    if self.valuesPerPoint > 1:
      return self.__consolidatingGenerator( list.__iter__(self) )
    else:
      return list.__iter__(self)

  def consolidate(self, valuesPerPoint):
    self.valuesPerPoint = int(valuesPerPoint)

  __consolidation_functions = {
    'sum': sum,
    'average': lambda usable: sum(usable) / len(usable),
    'avg_zero': lambda usable: sum(usable) / len(usable),
    'max': max,
    'min': min,
    'first': lambda usable: usable[0],
    'last': lambda usable: usable[-1],
  }
  __consolidation_function_aliases = {
    'avg': 'average',
  }

  def __consolidatingGenerator(self, gen):
    if self.consolidationFunc in self.__consolidation_functions:
      cf = self.__consolidation_functions[self.consolidationFunc]
    elif self.consolidationFunc in self.__consolidation_function_aliases:
      cf = self.__consolidation_functions[self.__consolidation_function_aliases[self.consolidationFunc]]
    else:
      raise Exception("Invalid consolidation function: '%s'" % self.consolidationFunc)

    buf = []
    valcnt = 0
    nonNull = 0

    for x in gen:
      valcnt += 1
      if x is not None:
        buf.append(x)
        nonNull += 1
      elif self.consolidationFunc == 'avg_zero':
        buf.append(0)

      if valcnt == self.valuesPerPoint:
        if nonNull and (nonNull / self.valuesPerPoint) >= self.xFilesFactor:
          yield cf(buf)
        else:
          yield None
        buf = []
        valcnt = 0
        nonNull = 0

    if valcnt > 0:
      if nonNull and (nonNull / self.valuesPerPoint) >= self.xFilesFactor:
        yield cf(buf)
      else:
        yield None

    return

  def __repr__(self):
    return 'TimeSeries(name=%s, start=%s, end=%s, step=%s, valuesPerPoint=%s, consolidationFunc=%s, xFilesFactor=%s)' % (
      self.name, self.start, self.end, self.step, self.valuesPerPoint, self.consolidationFunc, self.xFilesFactor)

  def getInfo(self):
    """Pickle-friendly representation of the series"""
    # make sure everything is unicode in python 2.x and 3.x
    return {
      text_type('name') : text_type(self.name),
      text_type('start') : self.start,
      text_type('end') : self.end,
      text_type('step') : self.step,
      text_type('values') : list(self),
      text_type('pathExpression') : text_type(self.pathExpression),
      text_type('valuesPerPoint') : self.valuesPerPoint,
      text_type('consolidationFunc'): text_type(self.consolidationFunc),
      text_type('xFilesFactor') : self.xFilesFactor,
    }

  def copy(self, name=None, start=None, end=None, step=None, values=None, consolidate=None, tags=None, xFilesFactor=None):
    return TimeSeries(
      name if name is not None else self.name,
      start if start is not None else self.start,
      end if end is not None else self.end,
      step if step is not None else self.step,
      values if values is not None else self.values,
      consolidate=consolidate if consolidate is not None else self.consolidationFunc,
      tags=tags if tags is not None else self.tags,
      xFilesFactor=xFilesFactor if xFilesFactor is not None else self.xFilesFactor
    )

  def datapoints(self):
    timestamps = range(int(self.start), int(self.end) + 1, int(self.step * self.valuesPerPoint))
    return list(zip(self, timestamps))

  @property
  def tags(self):
    return self.__tags

  @tags.setter
  def tags(self, tags):
    if isinstance(tags, Tags):
      self.__tags = tags
    elif isinstance(tags, dict):
      self.__tags = Tags(tags)
    else:
      raise Exception('Invalid tags specified')


# Data retrieval API
@logtime
def fetchData(requestContext, pathExpr, timer=None):
  timer.set_msg("lookup and merge of \"%s\" took" % str(pathExpr))

  seriesList = {}
  (startTime, endTime, now) = timebounds(requestContext)

  prefetched = requestContext.get('prefetched', {}).get((startTime, endTime, now), {}).get(pathExpr)
  if not prefetched:
    return []

  return _merge_results(pathExpr, startTime, endTime, prefetched, seriesList, requestContext)


def _merge_results(pathExpr, startTime, endTime, prefetched, seriesList, requestContext):
  log.debug("render.datalib.fetchData :: starting to merge")

  # Used as a cache to avoid recounting series None values below.
  series_best_nones = {}

  for path, results in prefetched:
    if not results:
      log.debug("render.datalib.fetchData :: no results for %s.fetch(%s, %s)" % (path, startTime, endTime))
      continue

    try:
      (timeInfo, values) = results
    except ValueError as e:
      raise Exception("could not parse timeInfo/values from metric '%s': %s" % (path, e))
    (start, end, step) = timeInfo

    series = TimeSeries(path, start, end, step, values, xFilesFactor=requestContext.get('xFilesFactor'))

    # hack to pass expressions through to render functions
    series.pathExpression = pathExpr

    if series.name in seriesList:
      # This counts the Nones in each series, and is unfortunately O(n) for each
      # series, which may be worth further optimization. The value of doing this
      # at all is to avoid the "flipping" effect of loading a graph multiple times
      # and having inconsistent data returned if one of the backing stores has
      # inconsistent data. This is imperfect as a validity test, but in practice
      # nicely keeps us using the "most complete" dataset available. Think of it
      # as a very weak CRDT resolver.
      candidate_nones = 0
      if not settings.REMOTE_STORE_MERGE_RESULTS:
        candidate_nones = len(
          [val for val in values if val is None])

      known = seriesList[series.name]
      # To avoid repeatedly recounting the 'Nones' in series we've already seen,
      # cache the best known count so far in a dict.
      if known.name in series_best_nones:
        known_nones = series_best_nones[known.name]
      else:
        known_nones = len([val for val in known if val is None])
        series_best_nones[known.name] = known_nones

      if known_nones > candidate_nones and len(series):
        if settings.REMOTE_STORE_MERGE_RESULTS and len(series) == len(known):
          # This series has potential data that might be missing from
          # earlier series.  Attempt to merge in useful data and update
          # the cache count.
          log.debug("Merging multiple TimeSeries for %s" % known.name)
          for i, j in enumerate(known):
            if j is None and series[i] is not None:
              known[i] = series[i]
              known_nones -= 1
          # Store known_nones in our cache
          series_best_nones[known.name] = known_nones
        else:
          # Not merging data -
          # we've found a series better than what we've already seen. Update
          # the count cache and replace the given series in the array.
          series_best_nones[known.name] = candidate_nones
          seriesList[known.name] = series

    else:
      # If we looked at this series above, and it matched a 'known'
      # series already, then it's already in the series list (or ignored).
      # If not, append it here.
      seriesList[series.name] = series

  # Stabilize the order of the results by ordering the resulting series by name.
  # This returns the result ordering to the behavior observed pre PR#1010.
  return [seriesList[k] for k in sorted(seriesList)]


def prefetchData(requestContext, pathExpressions):
  """Prefetch a bunch of path expressions and stores them in the context.

  The idea is that this will allow more batching than doing a query
  each time evaluateTarget() needs to fetch a path. All the prefetched
  data is stored in the requestContext, to be accessed later by fetchData.
  """
  if not pathExpressions:
    return

  start = time.time()
  log.debug("Fetching data for [%s]" % (', '.join(pathExpressions)))

  (startTime, endTime, now) = timebounds(requestContext)

  prefetched = collections.defaultdict(list)

  for result in STORE.fetch(pathExpressions, startTime, endTime, now, requestContext):
    if result is None:
      continue

    prefetched[result['pathExpression']].append((
      result['name'],
      (
        result['time_info'],
        result['values'],
      ),
    ))

  # Several third-party readers including rrdtool and biggraphite return values in a
  # generator which can only be iterated on once. These must be converted to a list.
  for pathExpression, items in prefetched.items():
    for i, (name, (time_info, values)) in enumerate(items):
      if isinstance(values, types.GeneratorType):
        prefetched[pathExpression][i] = (name, (time_info, list(values)))

  if not requestContext.get('prefetched'):
    requestContext['prefetched'] = {}

  if (startTime, endTime, now) in requestContext['prefetched']:
    for k in prefetched:
      requestContext['prefetched'][(startTime, endTime, now)][k] = prefetched[k]
  else:
      requestContext['prefetched'][(startTime, endTime, now)] = prefetched

  log.rendering("Fetched data for [%s] in %fs" % (', '.join(pathExpressions), time.time() - start))



try:
    from django.urls import reverse
except ImportError:  # Django < 1.10
    from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.http import HttpResponse, HttpResponsePermanentRedirect

from graphite.url_shortener.baseconv import base62
from graphite.url_shortener.models import Link
import re


def follow(request, link_id):
    """Follow existing links"""
    key = base62.to_decimal(link_id)
    link = get_object_or_404(Link, pk=key)
    return HttpResponsePermanentRedirect(reverse('browser') + link.url)


def shorten(request, path):
    if request.META.get('QUERY_STRING', None):
        path += '?' + request.META['QUERY_STRING']
    # Remove _salt, _dc and _uniq to avoid creating many copies of the same URL
    path = re.sub('&_(uniq|salt|dc)=[0-9.]+', "", path)

    link, created = Link.objects.get_or_create(url=path)
    link_id = base62.from_decimal(link.id)
    url = reverse('follow', kwargs={'link_id': link_id})
    return HttpResponse(url, content_type='text/plain')



from django.db import models


class Series(models.Model):
    hash = models.CharField(max_length=64, unique=True)
    path = models.TextField()
    __str__ = lambda self: "Series [%s]" % self.path


class Tag(models.Model):
    tag = models.CharField(max_length=191, unique=True)
    __str__ = lambda self: "Tag [%s]" % self.tag


class TagValue(models.Model):
    value = models.CharField(max_length=191, unique=True)
    __str__ = lambda self: "TagValue [%s]" % self.value


class SeriesTag(models.Model):
    series = models.ForeignKey(Series, on_delete=models.CASCADE)
    tag = models.ForeignKey(Tag, on_delete=models.CASCADE)
    value = models.ForeignKey(TagValue, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("series", "tag")

    __str__ = lambda self: "SeriesTag [%s %s %s]" % (self.series, self.tag, self.value)



import json

from graphite.util import jsonResponse, HttpResponse, HttpError
from graphite.functions import SeriesFunctions, SeriesFunction, PieFunctions, PieFunction, functionInfo


class jsonInfinityEncoder(json.JSONEncoder):
    def encode(self, o):
        return super(jsonInfinityEncoder, self).encode(o).replace('Infinity,', '1e9999,').replace('Infinity}', '1e9999}')

    def default(self, o):
        if hasattr(o, 'toJSON'):
            return o.toJSON()
        return o.__dict__


@jsonResponse(encoder=jsonInfinityEncoder)
def functionList(request, queryParams):
    if request.method != 'GET':
        return HttpResponse(status=405)

    if queryParams.get('type') == 'pie':
        funcs = PieFunctions()
    else:
        funcs = SeriesFunctions()

    grouped = queryParams.get('grouped', '').lower() in ['1', 'true']
    group = queryParams.get('group')
    result = {}

    for (name, func) in funcs.items():
        info = functionInfo(name, func)
        if group is not None and group != info['group']:
            continue

        if grouped:
            if info['group'] not in result:
                result[info['group']] = {}
            result[info['group']][name] = info
        else:
            result[name] = info

    return result


@jsonResponse(encoder=jsonInfinityEncoder)
def functionDetails(request, queryParams, name):
    if request.method != 'GET':
        return HttpResponse(status=405)

    try:
        if queryParams.get('type') == 'pie':
            func = PieFunction(name)
        else:
            func = SeriesFunction(name)
    except KeyError:
        raise HttpError('Function not found: %s' % name, status=404)

    return functionInfo(name, func)



# From https://github.com/reyoung/singleton
from threading import Lock


class Singleton(object):
    """
    The Singleton class decorator.
    Like:
        from singleton.singleton import Singleton

        @Singleton
        class IntSingleton(object):
            def __init__(self):
                pass
    Use IntSingleton.instance() get the instance
    """

    def __init__(self, cls):
        """
        :param cls: decorator class type
        """
        self.__cls = cls
        self.__instance = None

    def initialize(self, *args, **kwargs):
        """
        Initialize singleton object if it has not been initialized
        :param args: class init parameters
        :param kwargs: class init parameters
        """
        if not self.is_initialized():
            self.__instance = self.__cls(*args, **kwargs)

    def is_initialized(self):
        """
        :return: true if instance is initialized
        """
        return self.__instance is not None

    def instance(self):
        """
        Get singleton instance
        :return: instance object
        """
        if not self.is_initialized():
            self.initialize()
        return self.__instance

    def __call__(self, *args, **kwargs):
        """
        Disable new instance of original class
        :raise TypeError:
        """
        raise TypeError("Singletons must be access by instance")

    def __instancecheck__(self, inst):
        """
        Helper for isinstance check
        """
        return isinstance(inst, self.__cls)


class ThreadSafeSingleton(object):
    def __init__(self, cls):
        self.__cls = cls
        self.__instance = None
        self.__mutex = Lock()

    def is_initialized(self):
        self.__mutex.acquire()
        try:
            return self.__instance is not None
        finally:
            self.__mutex.release()

    def initialize(self, *args, **kwargs):
        self.__mutex.acquire()
        try:
            if self.__instance is None:
                self.__instance = self.__cls(*args, **kwargs)
        finally:
            self.__mutex.release()

    def instance(self):
        self.__mutex.acquire()
        try:
            if self.__instance is None:
                self.__instance = self.__cls()
            return self.__instance
        finally:
            self.__mutex.release()

    def __call__(self, *args, **kwargs):
        """
        Disable new instance of original class
        :raise TypeError:
        """
        raise TypeError("Singletons must be access by instance")

    def __instancecheck__(self, inst):
        """
        Helper for isinstance check
        """
        return isinstance(inst, self.__cls)



from __future__ import absolute_import
import time

# Use the built-in version of scandir/stat if possible, otherwise
# use the scandir module version
try:
    from os import scandir, stat  # noqa # pylint: disable=unused-import
except ImportError:
    from scandir import scandir, stat  # noqa # pylint: disable=unused-import

try:
    import whisper
except ImportError:
    whisper = False

try:
    import gzip
except ImportError:
    gzip = False


from graphite.intervals import Interval, IntervalSet
from graphite.logger import log
from graphite.readers.utils import merge_with_carbonlink, BaseReader

# The parser was replacing __readHeader with the <class>__readHeader
# which was not working.
if bool(whisper):
    whisper__readHeader = whisper.__readHeader


class WhisperReader(BaseReader):
    __slots__ = ('fs_path', 'real_metric_path')
    supported = bool(whisper)
    meta_info = None

    def __init__(self, fs_path, real_metric_path):
        self.fs_path = fs_path
        self.real_metric_path = real_metric_path

    def info(self):
        if not self.meta_info:
            self.meta_info = whisper.info(self.fs_path)
        return self.meta_info

    def get_raw_step(self):
        return self.info()['archives'][0]['secondsPerPoint']

    def get_intervals(self):
        start = time.time() - self.info()['maxRetention']
        end = max(stat(self.fs_path).st_mtime, start)
        return IntervalSet([Interval(start, end)])

    def fetch_data(self, startTime, endTime, now=None):
        return whisper.fetch(self.fs_path, startTime, endTime, now=now)

    def fetch(self, startTime, endTime, now=None, requestContext=None):
        try:
            data = self.fetch_data(startTime, endTime, now=now)
        except IOError:
            log.exception("Failed fetch of whisper file '%s'" % self.fs_path)
            return None
        if not data:
            return None

        time_info, values = data
        (start, end, step) = time_info

        meta_info = self.info()
        aggregation_method = meta_info['aggregationMethod']

        # Merge in data from carbon's cache
        values = merge_with_carbonlink(
            self.real_metric_path, start, step, values, aggregation_method, self.get_raw_step())

        return time_info, values


class GzippedWhisperReader(WhisperReader):
    supported = bool(whisper and gzip)

    def info(self):
        if not self.meta_info:
            fh = gzip.GzipFile(self.fs_path, 'rb')
            try:
                self.meta_info = whisper__readHeader(fh)  # evil, but necessary.
            finally:
                fh.close()
        return self.meta_info

    def fetch_data(self, startTime, endTime, now=None):
        fh = gzip.GzipFile(self.fs_path, 'rb')
        try:
            return whisper.file_fetch(fh, startTime, endTime, now=now)
        finally:
            fh.close()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-14 11:22
from __future__ import unicode_literals

from django.db import migrations, models
import tagging.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('when', models.DateTimeField()),
                ('what', models.CharField(max_length=255)),
                ('data', models.TextField(blank=True)),
                ('tags', tagging.fields.TagField(blank=True, default=b'', max_length=255)),
            ],
        ),
    ]





class Node(object):
    __slots__ = ('name', 'path', 'local', 'is_leaf')

    def __init__(self, path):
        self.path = path
        self.name = path.split('.')[-1]
        self.local = True
        self.is_leaf = False

    def __repr__(self):
        return '<%s[%x]: %s>' % (self.__class__.__name__, id(self), self.path)


class BranchNode(Node):
    pass


class LeafNode(Node):
    __slots__ = ('reader', )

    def __init__(self, path, reader):
        Node.__init__(self, path)
        self.reader = reader
        self.is_leaf = True

    def fetch(self, startTime, endTime, now=None, requestContext=None):
        try:
            result = self.reader.fetch(startTime, endTime, now, requestContext)
        except TypeError:
            # Support for legacy 3rd party, readers.
            result = self.reader.fetch(startTime, endTime)

        return result

    @property
    def intervals(self):
        return self.reader.get_intervals()

    def __repr__(self):
        return '<LeafNode[%x]: %s (%s)>' % (id(self), self.path, self.reader)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import traceback

import ldap
from django.conf import settings
from django.contrib.auth.models import User


class LDAPBackend:
    def authenticate(self, username=None, password=None):
        if settings.LDAP_USER_DN_TEMPLATE is not None:
            settings.LDAP_BASE_USER = settings.LDAP_USER_DN_TEMPLATE % {'username': username}
            settings.LDAP_BASE_PASS = password
        try:
            conn = ldap.initialize(settings.LDAP_URI)
            conn.protocol_version = ldap.VERSION3
            if settings.LDAP_USE_TLS:
                conn.start_tls_s()
            conn.simple_bind_s( settings.LDAP_BASE_USER, settings.LDAP_BASE_PASS )
        except ldap.LDAPError:
            traceback.print_exc()
            return None

        scope = ldap.SCOPE_SUBTREE
        filter = settings.LDAP_USER_QUERY % username
        returnFields = ['dn','mail']
        try:
            resultID = conn.search( settings.LDAP_SEARCH_BASE, scope, filter, returnFields )
            resultType, resultData = conn.result( resultID, 0 )
            if len(resultData) != 1: # User does not exist
                return None

            userDN = resultData[0][0]
            try:
                userMail = resultData[0][1]['mail'][0].decode("utf-8")
            except Exception:
                userMail = "Unknown"

            conn.simple_bind_s(userDN,password)
            try:
                user = User.objects.get(username=username)
            except Exception:  # First time login, not in django's database
                # To prevent login from django db user
                randomPasswd = User.objects.make_random_password(length=16)
                user = User.objects.create_user(username, userMail, randomPasswd)
                user.save()

            return user

        except ldap.INVALID_CREDENTIALS:
            traceback.print_exc()
            return None

    def get_user(self,user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



from django.shortcuts import render
from graphite import settings


def index(request):
    context = {
        'version' : settings.WEBAPP_VERSION,
    }
    return render(request, 'version.html', context)



from __future__ import division

import abc

from graphite.logger import log


class BaseReader(object):
    __metaclass__ = abc.ABCMeta

    supported = True

    @abc.abstractmethod
    def get_intervals(self):
        """Get the supported interval by a reader.

        Returns:
          IntervalSet(): set of supported intervals.
        """

    @abc.abstractmethod
    def fetch(self, startTime, endTime, now=None, requestContext=None):
        """Fetches points for a given interval.

        Args:
          startTime: int
          endTime: int
          now: int
          requestContext: RequestContext

        Returns:
          (time_info, values)
        """


def merge_with_cache(cached_datapoints, start, step, values, func=None, raw_step=None):
    """Merge values with datapoints from a buffer/cache."""
    consolidated = []

    # Similar to the function in render/datalib:TimeSeries
    def consolidate(func, values):
        usable = [v for v in values if v is not None]
        if not usable:
            return None
        if func == 'avg_zero':
            return sum([0 if v is None else v for v in values]) / len(values)
        if func == 'sum':
            return sum(usable)
        if func == 'average':
            return sum(usable) / len(usable)
        if func == 'max':
            return max(usable)
        if func == 'min':
            return min(usable)
        if func == 'last':
            return usable[-1]
        raise Exception("Invalid consolidation function: '%s'" % func)

    # if we have a raw_step, start by taking only the last data point for each interval to match what whisper will do
    if raw_step is not None and raw_step > 1:
        consolidated_dict = {}
        for (timestamp, value) in cached_datapoints:
            interval = timestamp - (timestamp % raw_step)
            consolidated_dict[interval] = value
        cached_datapoints = list(consolidated_dict.items())

    # if we have a consolidation function and the step is not the default interval, consolidate to the requested step
    if func and step != raw_step:
        consolidated_dict = {}
        for (timestamp, value) in cached_datapoints:
            interval = timestamp - (timestamp % step)
            if interval in consolidated_dict:
                consolidated_dict[interval].append(value)
            else:
                consolidated_dict[interval] = [value]
        consolidated = [(i, consolidate(func, consolidated_dict[i])) for i in consolidated_dict]
    # otherwise just use the points
    else:
        consolidated = cached_datapoints

    for (interval, value) in consolidated:
        try:
            i = int(interval - start) // step
            if i < 0:
                # cached data point is earlier then the requested data point.
                # meaning we can definitely ignore the cache result.
                # note that we cannot rely on the 'except'
                # in this case since 'values[-n]='
                # is equivalent to 'values[len(values) - n]='
                continue
            values[i] = value
        except BaseException:
            pass

    return values


def CarbonLink():
    """Return a carbonlink instance."""
    # Late import to avoid pulling out too many dependencies with
    # readers.py which is usually imported by plugins.
    from graphite.carbonlink import CarbonLink
    return CarbonLink()


def merge_with_carbonlink(metric, start, step, values, aggregation_method=None, raw_step=None):
    """Get points from carbonlink and merge them with existing values."""
    cached_datapoints = []
    try:
        cached_datapoints = CarbonLink().query(metric)
    except BaseException:
        log.exception("Failed CarbonLink query '%s'" % metric)
        cached_datapoints = []

    if isinstance(cached_datapoints, dict):
        cached_datapoints = list(cached_datapoints.items())

    return merge_with_cache(
        cached_datapoints, start, step, values,
        func=aggregation_method, raw_step=raw_step)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import re

from django.conf import settings
from django.shortcuts import render
from django.utils.safestring import mark_safe
from django.utils.html import escape
from graphite.account.models import Profile
from graphite.compat import HttpResponse
from graphite.user_util import getProfile, getProfileByUsername
from graphite.util import json
from graphite.logger import log
from hashlib import md5


def header(request):
  "View for the header frame of the browser UI"
  context = {}
  context['user'] = request.user
  context['profile'] = getProfile(request)
  context['documentation_url'] = settings.DOCUMENTATION_URL
  context['login_url'] = settings.LOGIN_URL
  return render(request, "browserHeader.html", context)


def browser(request):
  "View for the top-level frame of the browser UI"
  context = {
    'queryString': mark_safe(request.GET.urlencode()),
    'target': request.GET.get('target')
  }
  if context['queryString']:
    context['queryString'] = context['queryString'].replace('#','%23')
  if context['target']:
    context['target'] = context['target'].replace('#','%23') #js libs terminate a querystring on #
  return render(request, "browser.html", context)


def search(request):
  query = request.POST.get('query')
  if not query:
    return HttpResponse("")

  patterns = query.split()
  regexes = [re.compile(p,re.I) for p in patterns]

  def matches(s):
    for regex in regexes:
      if regex.search(s):
        return True
    return False

  results = []

  index_file = open(settings.INDEX_FILE)
  for line in index_file:
    if matches(line):
      results.append( line.strip() )
    if len(results) >= 100:
      break

  index_file.close()
  result_string = ','.join(results)
  return HttpResponse(result_string, content_type='text/plain')


def myGraphLookup(request):
  "View for My Graphs navigation"
  profile = getProfile(request,allowDefault=False)
  assert profile

  nodes = []
  leafNode = {
    'allowChildren' : 0,
    'expandable' : 0,
    'leaf' : 1,
  }
  branchNode = {
    'allowChildren' : 1,
    'expandable' : 1,
    'leaf' : 0,
  }

  try:
    path = request.GET.get('path', u'')

    if path:
      if path.endswith('.'):
        userpath_prefix = path

      else:
        userpath_prefix = path + '.'

    else:
      userpath_prefix = u""

    matches = [ graph for graph in profile.mygraph_set.all().order_by('name') if graph.name.startswith(userpath_prefix) ]

    log.info( "myGraphLookup: username=%s, path=%s, userpath_prefix=%s, %ld graph to process" % (profile.user.username, path, userpath_prefix, len(matches)) )
    branch_inserted = set()
    leaf_inserted = set()

    for graph in matches: #Now let's add the matching graph
      isBranch = False
      dotPos = graph.name.find( '.', len(userpath_prefix) )

      if dotPos >= 0:
        isBranch = True
        name = graph.name[ len(userpath_prefix) : dotPos ]
        if name in branch_inserted: continue
        branch_inserted.add(name)

      else:
         name = graph.name[ len(userpath_prefix): ]
         if name in leaf_inserted: continue
         leaf_inserted.add(name)

      node = {'text': escape(name)}

      if isBranch:
        node.update({'id': userpath_prefix + name + '.'})
        node.update(branchNode)

      else:
        m = md5()
        m.update(name.encode('utf-8'))
        node.update( { 'id' : str(userpath_prefix + m.hexdigest()), 'graphUrl' : graph.url } )
        node.update(leafNode)

      nodes.append(node)

  except Exception:
    log.exception("browser.views.myGraphLookup(): could not complete request.")

  if not nodes:
    no_graphs = { 'text' : "No saved graphs", 'id' : 'no-click' }
    no_graphs.update(leafNode)
    nodes.append(no_graphs)

  return json_response(nodes, request)


def userGraphLookup(request):
  "View for User Graphs navigation"
  user = request.GET.get('user')
  path = request.GET['path']

  if user:
    username = user
    graphPath = path[len(username)+1:]
  elif '.' in path:
    username, graphPath = path.split('.', 1)
  else:
    username, graphPath = path, None

  nodes = []

  branchNode = {
    'allowChildren' : 1,
    'expandable' : 1,
    'leaf' : 0,
  }
  leafNode = {
    'allowChildren' : 0,
    'expandable' : 0,
    'leaf' : 1,
  }

  try:

    if not username:
      profiles = Profile.objects.exclude(user__username='default').order_by('user__username')

      for profile in profiles:
        if profile.mygraph_set.count():
          node = {
            'text' : profile.user.username,
            'id' : profile.user.username,
          }

          node.update(branchNode)
          nodes.append(node)

    else:
      profile = getProfileByUsername(username)
      assert profile, "No profile for username '%s'" % username

      if graphPath:
        prefix = graphPath.rstrip('.') + '.'
      else:
        prefix = ''

      matches = [ graph for graph in profile.mygraph_set.order_by('name') if graph.name.startswith(prefix) ]
      inserted = set()

      for graph in matches:
        relativePath = graph.name[ len(prefix): ]
        nodeName = relativePath.split('.')[0]

        if nodeName in inserted:
          continue
        inserted.add(nodeName)

        if '.' in relativePath: # branch
          node = {
            'text' : escape(nodeName),
            'id' : username + '.' + prefix + nodeName + '.',
          }
          node.update(branchNode)
        else: # leaf
          m = md5()
          m.update(nodeName.encode('utf-8'))

          node = {
            'text' : escape(nodeName),
            'id' : username + '.' + prefix + m.hexdigest(),
            'graphUrl' : graph.url,
          }
          node.update(leafNode)

        nodes.append(node)

  except Exception:
    log.exception("browser.views.userLookup(): could not complete request for %s" % username)

  if not nodes:
    no_graphs = { 'text' : "No saved graphs", 'id' : 'no-click' }
    no_graphs.update(leafNode)
    nodes.append(no_graphs)

  nodes.sort(key=lambda node: node['allowChildren'], reverse = True)

  return json_response(nodes, request)


def json_response(nodes, request=None):
  if request:
    jsonp = request.GET.get('jsonp', False) or request.POST.get('jsonp', False)
  else:
    jsonp = False
  #json = str(nodes) #poor man's json encoder for simple types
  json_data = json.dumps(nodes)
  if jsonp:
    response = HttpResponse("%s(%s)" % (jsonp, json_data),
                            content_type="text/javascript")
  else:
    response = HttpResponse(json_data, content_type="application/json")
  response['Pragma'] = 'no-cache'
  response['Cache-Control'] = 'no-cache'
  return response



import fnmatch
import os.path
import re

EXPAND_BRACES_RE = re.compile(r'(\{([^\{\}]*)\})')


def get_real_metric_path(absolute_path, metric_path):
    # Support symbolic links (real_metric_path ensures proper cache queries)
    real_absolute_path = os.path.realpath(absolute_path)
    if absolute_path != real_absolute_path:
        # replace left side base_fs_path that contains sym link with real fs path
        relative_fs_path = metric_path.replace('.', os.sep)
        absolute_path_no_ext, _ext = os.path.splitext(absolute_path)
        base_fs_path = os.path.dirname(absolute_path_no_ext[:-len(relative_fs_path)])
        real_base_fs_path = os.path.realpath(base_fs_path)
        real_relative_fs_path = real_absolute_path[len(real_base_fs_path):].lstrip(os.sep)
        return fs_to_metric(real_relative_fs_path)

    return metric_path


def fs_to_metric(path):
    dirpath = os.path.dirname(path)
    filename = os.path.basename(path)
    return os.path.join(dirpath, filename.split('.')[0]).replace(os.sep, '.')


def extract_variants(pattern):
    """Extract the pattern variants (ie. {foo,bar}baz = foobaz or barbaz)."""
    v1, v2 = pattern.find('{'), pattern.find('}')
    if v1 > -1 and v2 > v1:
        variations = pattern[v1 + 1:v2].split(',')
        variants = [pattern[:v1] + v + pattern[v2 + 1:] for v in variations]
        return list({r for v in variants for r in extract_variants(v)})
    else:
        return [pattern]


def match_entries(entries, pattern):
    # First we check for pattern variants (ie. {foo,bar}baz = foobaz or barbaz)
    matching = []

    for variant in expand_braces(pattern):
        matching.extend(fnmatch.filter(entries, variant))

    return list(set(matching))


"""
  Brace expanding patch for python3 borrowed from:
  https://bugs.python.org/issue9584
"""


def expand_braces(s):
    res = list()

    m = EXPAND_BRACES_RE.search(s)
    if m is not None:
        sub = m.group(2)
        open_brace, close_brace = m.span(1)
        for pat in sub.split(','):
            res.extend(expand_braces(s[:open_brace] + pat + s[close_brace:]))
    else:
        res.append(s)

    return list(set(res))



from graphite.errors import InputParameterError
from graphite.functions import safe


aggFuncs = {
  'average': safe.safeAvg,
  'avg_zero': safe.safeAvgZero,
  'median': safe.safeMedian,
  'sum': safe.safeSum,
  'min': safe.safeMin,
  'max': safe.safeMax,
  'diff': safe.safeDiff,
  'stddev': safe.safeStdDev,
  'count': safe.safeLen,
  'range': lambda row: safe.safeSubtract(safe.safeMax(row), safe.safeMin(row)),
  'multiply': lambda row: safe.safeMul(*row),
  'last': safe.safeLast,
}


aggFuncAliases = {
  'rangeOf': aggFuncs['range'],
  'avg': aggFuncs['average'],
  'total': aggFuncs['sum'],
  'current': aggFuncs['last'],
}


def getAggFunc(func, rawFunc=None):
    if func in aggFuncs:
        return aggFuncs[func]
    if func in aggFuncAliases:
        return aggFuncAliases[func]
    raise InputParameterError('Unsupported aggregation function: %s' % (rawFunc or func))



from __future__ import absolute_import

import re
import bisect
import sys

from graphite.tags.base import BaseTagDB, TaggedSeries


class RedisTagDB(BaseTagDB):
    """
    Stores tag information in a Redis database.

    Keys used are:

    .. code-block:: none

      series                    # Set of all paths
      series:<path>:tags        # Hash of all tag:value pairs for path
      tags                      # Set of all tags
      tags:<tag>:series         # Set of paths with entry for tag
      tags:<tag>:values         # Set of values for tag
      tags:<tag>:values:<value> # Set of paths matching tag/value

    """
    def __init__(self, settings, *args, **kwargs):
        super(RedisTagDB, self).__init__(settings, *args, **kwargs)

        from redis import Redis

        self.r = Redis(
            host=settings.TAGDB_REDIS_HOST,
            port=settings.TAGDB_REDIS_PORT,
            db=settings.TAGDB_REDIS_DB,
            password=settings.TAGDB_REDIS_PASSWORD,
            decode_responses=(sys.version_info[0] >= 3),
        )

    def _find_series(self, tags, requestContext=None):
        selector = None
        selector_cnt = None
        filters = []

        # loop through tagspecs, look for best spec to use as selector
        for tagspec in tags:
            (tag, operator, spec) = self.parse_tagspec(tagspec)

            if operator == '=':
                matches_empty = spec == ''
                if not matches_empty:
                    cnt = self.r.scard('tags:' + tag + ':values:' + spec)
                    if not selector or selector[1] != '=' or selector_cnt > cnt:
                        if selector:
                            filters.append(selector)
                        selector = (tag, operator, spec)
                        selector_cnt = cnt
                        continue
                filters.append((tag, operator, spec))

            elif operator == '=~':
                pattern = re.compile(spec)
                matches_empty = bool(pattern.match(''))
                if not matches_empty and (not selector or selector[1] != '='):
                    cnt = self.r.scard('tags:' + tag + ':values')
                    if not selector or selector_cnt > cnt:
                        if selector:
                            filters.append(selector)
                        selector = (tag, operator, pattern)
                        selector_cnt = cnt
                        continue
                filters.append((tag, operator, pattern))

            elif operator == '!=':
                matches_empty = spec != ''
                if not matches_empty and (not selector or selector[1] != '='):
                    cnt = self.r.scard('tags:' + tag + ':values')
                    if not selector or selector_cnt > cnt:
                        if selector:
                            filters.append(selector)
                        selector = (tag, operator, spec)
                        selector_cnt = cnt
                        continue
                filters.append((tag, operator, spec))

            elif operator == '!=~':
                pattern = re.compile(spec)
                matches_empty = not pattern.match('')
                if not matches_empty and (not selector or selector[1] != '='):
                    cnt = self.r.scard('tags:' + tag + ':values')
                    if not selector or selector_cnt > cnt:
                        if selector:
                            filters.append(selector)
                        selector = (tag, operator, pattern)
                        selector_cnt = cnt
                        continue
                filters.append((tag, operator, pattern))

            else:
                raise ValueError("Invalid operator %s" % operator)

        if not selector:
            raise ValueError("At least one tagspec must not match the empty string")

        # get initial list of series
        (tag, operator, spec) = selector

        # find list of values that match the tagspec
        values = None
        if operator == '=':
            values = [spec]
        elif operator == '=~':
            # see if we can identify a literal prefix to filter by in redis
            match = None
            m = re.match('([a-z0-9]+)([^*?|][^|]*)?$', spec.pattern)
            if m:
                match = m.group(1) + '*'
            values = [value for value in self.r.sscan_iter('tags:' + tag + ':values', match=match) if spec.match(value) is not None]
        elif operator == '!=':
            values = [value for value in self.r.sscan_iter('tags:' + tag + ':values') if value != spec]
        elif operator == '!=~':
            values = [value for value in self.r.sscan_iter('tags:' + tag + ':values') if spec.match(value) is None]

        # if this query matched no values, just short-circuit since the result of the final intersect will be empty
        if not values:
            return []

        results = []

        # apply filters
        operators = ['=','!=','=~','!=~']
        filters.sort(key=lambda a: operators.index(a[1]))

        for series in self.r.sunion(*['tags:' + tag + ':values:' + value for value in values]):
            try:
                parsed = self.parse(series)
            except Exception:
                continue

            matched = True

            for (tag, operator, spec) in filters:
                value = parsed.tags.get(tag, '')
                if (
                    (operator == '=' and value != spec) or
                    (operator == '=~' and spec.match(value) is None) or
                    (operator == '!=' and value == spec) or
                    (operator == '!=~' and spec.match(value) is not None)
                ):
                    matched = False
                    break

            if matched:
                bisect.insort_left(results, series)

        return results

    def get_series(self, path, requestContext=None):
        tags = {}

        tags = self.r.hgetall('series:' + path + ':tags')
        if not tags:
            return None

        return TaggedSeries(tags['name'], tags)

    def list_tags(self, tagFilter=None, limit=None, requestContext=None):
        result = []

        if tagFilter:
            tagFilter = re.compile(tagFilter)

        for tag in self.r.sscan_iter('tags'):
            if tagFilter and tagFilter.match(tag) is None:
                continue
            if len(result) == 0 or tag >= result[-1]:
                if limit and len(result) >= limit:
                    continue
                result.append(tag)
            else:
                bisect.insort_left(result, tag)
            if limit and len(result) > limit:
                del result[-1]

        return [
            {'tag': tag}
            for tag in result
        ]

    def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
        if not self.r.sismember('tags', tag):
            return None

        return {
            'tag': tag,
            'values': self.list_values(
                tag,
                valueFilter=valueFilter,
                limit=limit,
                requestContext=requestContext
            ),
        }

    def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
        result = []

        if valueFilter:
            valueFilter = re.compile(valueFilter)

        for value in self.r.sscan_iter('tags:' + tag + ':values'):
            if valueFilter and valueFilter.match(value) is None:
                continue
            if len(result) == 0 or value >= result[-1]:
                if limit and len(result) >= limit:
                    continue
                result.append(value)
            else:
                bisect.insort_left(result, value)
            if limit and len(result) > limit:
                del result[-1]

        return [
            {'value': value, 'count': self.r.scard('tags:' + tag + ':values:' + value)}
            for value in result
        ]

    def tag_series(self, series, requestContext=None):
        # extract tags and normalize path
        parsed = self.parse(series)

        path = parsed.path

        with self.r.pipeline() as pipe:
            pipe.sadd('series', path)

            for tag, value in parsed.tags.items():
                pipe.hset('series:' + path + ':tags', tag, value)

                pipe.sadd('tags', tag)
                pipe.sadd('tags:' + tag + ':series', path)
                pipe.sadd('tags:' + tag + ':values', value)
                pipe.sadd('tags:' + tag + ':values:' + value, path)

            pipe.execute()

        return path

    def del_series(self, series, requestContext=None):
        # extract tags and normalize path
        parsed = self.parse(series)

        path = parsed.path

        with self.r.pipeline() as pipe:
            pipe.srem('series', path)

            pipe.delete('series:' + path + ':tags')

            for tag, value in parsed.tags.items():
                pipe.srem('tags:' + tag + ':series', path)
                pipe.srem('tags:' + tag + ':values:' + value, path)

            pipe.execute()

        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-08-31 17:38
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Series',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hash', models.CharField(max_length=64, unique=True)),
                ('path', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='SeriesTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('series', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tags.Series')),
            ],
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tag', models.CharField(max_length=191, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='TagValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=191, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name='seriestag',
            name='tag',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tags.Tag'),
        ),
        migrations.AddField(
            model_name='seriestag',
            name='value',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tags.TagValue'),
        ),
        migrations.AlterUniqueTogether(
            name='seriestag',
            unique_together=set([('series', 'tag')]),
        ),
    ]



import os
import sys

try:
    from importlib import import_module
except ImportError:
    from django.utils.importlib import import_module

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'graphite.settings')  # noqa
os.environ.setdefault('GRAPHITE_SETTINGS_MODULE', 'graphite.local_settings')  # noqa

from django.conf import settings
from django.core.wsgi import get_wsgi_application

application = get_wsgi_application()

try:
    import whitenoise
except ImportError:
    whitenoise = False
else:
    whitenoise_version = tuple(map(
            int, getattr(whitenoise, '__version__', '0').split('.')))
    # WhiteNoise < 2.0.1 does not support Python 2.6
    if sys.version_info[:2] < (2, 7):
        if whitenoise_version < (2, 0, 1):
            whitenoise = False
    # Configure WhiteNoise >= 3.2 as middleware from app_settings.py
    # http://whitenoise.evans.io/en/stable/changelog.html#v4-0
    if whitenoise_version >= (3, 2):
        whitenoise = False

if whitenoise:
    from whitenoise.django import DjangoWhiteNoise
    application = DjangoWhiteNoise(application)
    prefix = "/".join((settings.URL_PREFIX.strip('/'), 'static'))
    for directory in settings.STATICFILES_DIRS:
        application.add_files(directory, prefix=prefix)
    for app_path in settings.INSTALLED_APPS:
        module = import_module(app_path)
        directory = os.path.join(os.path.dirname(module.__file__), 'static')
        if os.path.isdir(directory):
            application.add_files(directory, prefix=prefix)



import bisect
import fnmatch
import os
from os.path import isdir, isfile, join, basename, splitext
from django.conf import settings

# Use the built-in version of scandir/walk if possible, otherwise
# use the scandir module version
try:
    from os import scandir, walk
except ImportError:
    from scandir import scandir, walk

from graphite.logger import log
from graphite.node import BranchNode, LeafNode
from graphite.readers import WhisperReader, GzippedWhisperReader, RRDReader
from graphite.util import find_escaped_pattern_fields
from graphite.finders.utils import BaseFinder
from graphite.tags.utils import TaggedSeries

from . import fs_to_metric, get_real_metric_path, match_entries, expand_braces


class StandardFinder(BaseFinder):
    DATASOURCE_DELIMITER = '::RRD_DATASOURCE::'

    def __init__(self, directories=None):
        directories = directories or settings.STANDARD_DIRS
        self.directories = directories

    def find_nodes(self, query):
        clean_pattern = query.pattern.replace('\\', '')

        # translate query pattern if it is tagged
        tagged = not query.pattern.startswith('_tagged.') and ';' in query.pattern
        if tagged:
            # tagged series are stored in whisper using encoded names, so to retrieve them we need to
            # encode the query pattern using the same scheme used in carbon when they are written.
            encoded_paths = [
                TaggedSeries.encode(query.pattern, sep=os.sep, hash_only=True),
                TaggedSeries.encode(query.pattern, sep=os.sep, hash_only=False),
            ]

        pattern_parts = clean_pattern.split('.')

        for root_dir in self.directories:
            if tagged:
                relative_paths = []
                for pattern in encoded_paths:
                    entries = [
                        pattern + '.wsp',
                        pattern + '.wsp.gz',
                        pattern + '.rrd',
                    ]
                    for entry in entries:
                        if isfile(join(root_dir, entry)):
                            relative_paths.append(entry)
            else:
                relative_paths = self._find_paths(root_dir, pattern_parts)

            for relative_path in relative_paths:
                if basename(relative_path).startswith('.'):
                    continue

                if self.DATASOURCE_DELIMITER in basename(relative_path):
                    (relative_path, datasource_pattern) = relative_path.rsplit(
                        self.DATASOURCE_DELIMITER, 1)
                else:
                    datasource_pattern = None

                absolute_path = join(root_dir, relative_path)

                # if we're finding by tag, return the proper metric path
                if tagged:
                    metric_path = query.pattern
                    real_metric_path = query.pattern
                else:
                    metric_path = fs_to_metric(relative_path)
                    real_metric_path = get_real_metric_path(absolute_path, metric_path)
                    metric_path_parts = metric_path.split('.')
                    for field_index in find_escaped_pattern_fields(query.pattern):
                        metric_path_parts[field_index] = pattern_parts[field_index].replace('\\', '')
                    metric_path = '.'.join(metric_path_parts)

                # Now we construct and yield an appropriate Node object
                if isdir(absolute_path):
                    yield BranchNode(metric_path)

                elif absolute_path.endswith('.wsp') and WhisperReader.supported:
                    reader = WhisperReader(absolute_path, real_metric_path)
                    yield LeafNode(metric_path, reader)

                elif absolute_path.endswith('.wsp.gz') and GzippedWhisperReader.supported:
                    reader = GzippedWhisperReader(absolute_path, real_metric_path)
                    yield LeafNode(metric_path, reader)

                elif absolute_path.endswith('.rrd') and RRDReader.supported:
                    if datasource_pattern is None:
                        yield BranchNode(metric_path)

                    else:
                        for datasource_name in RRDReader.get_datasources(absolute_path):
                            if match_entries([datasource_name], datasource_pattern):
                                reader = RRDReader(absolute_path, datasource_name)
                                yield LeafNode(metric_path + "." + datasource_name, reader)

    def _find_paths(self, current_dir, patterns):
        """Recursively generates absolute paths whose components underneath current_dir
        match the corresponding pattern in patterns"""
        raw_pattern = patterns[0]
        patterns = patterns[1:]

        for pattern in expand_braces(raw_pattern):
            has_wildcard = pattern.find('[') > -1 or pattern.find('*') > -1 or pattern.find('?') > -1

            matching_subdirs = []
            files = []
            if has_wildcard:  # this avoids os.listdir() for performance
                subdirs = []
                try:
                    for x in scandir(current_dir):
                        if x.is_file():
                            files.append(x.name)
                        if x.is_dir():
                            subdirs.append(x.name)
                except OSError as e:
                    log.exception(e)

                if pattern == "**":
                    matching_subdirs = map(
                      lambda item: item[0][len(current_dir) + 1:],
                      walk(current_dir)
                    )

                    # if this is a terminal globstar, add a pattern for all files in subdirs
                    if not patterns:
                        patterns = ["*"]
                else:
                    matching_subdirs = match_entries(subdirs, pattern)
            elif isdir(join(current_dir, pattern)):
                matching_subdirs.append(pattern)

            # the last pattern may apply to RRD data sources
            if len(patterns) == 1 and RRDReader.supported:
                if not has_wildcard:
                    entries = [
                      pattern + ".rrd",
                    ]
                    rrd_files = [entry for entry in entries if isfile(join(current_dir, entry))]
                else:
                    rrd_files = match_entries(files, pattern + ".rrd")

                if rrd_files:  # let's assume it does
                    datasource_pattern = patterns[0]

                    for rrd_file in rrd_files:
                        yield rrd_file + self.DATASOURCE_DELIMITER + datasource_pattern

            if patterns:  # we've still got more directories to traverse
                for subdir in matching_subdirs:
                    absolute_path = join(current_dir, subdir)
                    for match in self._find_paths(absolute_path, patterns):
                        yield join(subdir, match)

            else:  # we've got the last pattern
                if not has_wildcard:
                    entries = [
                        pattern + '.wsp',
                        pattern + '.wsp.gz',
                        pattern + '.rrd',
                    ]
                    matching_files = [entry for entry in entries if isfile(join(current_dir, entry))]
                else:
                    matching_files = match_entries(files, pattern + '.*')

                for match in matching_files + matching_subdirs:
                    yield match

    def get_index(self, requestContext):
        matches = []

        for root, _, files in walk(settings.WHISPER_DIR):
            root = root.replace(settings.WHISPER_DIR, '')
            for base_name in files:
                if fnmatch.fnmatch(base_name, '*.wsp'):
                    match = join(root, base_name).replace('.wsp', '').replace('/', '.').lstrip('.')
                    bisect.insort_left(matches, match)

        # unlike 0.9.x, we're going to use os.walk with followlinks
        # since we require Python 2.7 and newer that supports it
        if RRDReader.supported:
            for root, _, files in walk(settings.RRD_DIR, followlinks=True):
                root = root.replace(settings.RRD_DIR, '')
                for base_name in files:
                    if fnmatch.fnmatch(base_name, '*.rrd'):
                        absolute_path = join(settings.RRD_DIR, root, base_name)
                        base_name = splitext(base_name)[0]
                        metric_path = join(root, base_name)
                        rrd = RRDReader(absolute_path, metric_path)
                        for datasource_name in rrd.get_datasources(absolute_path):
                            match = join(metric_path, datasource_name).replace('.rrd', '').replace('/', '.').lstrip('.')
                            if match not in matches:
                                bisect.insort_left(matches, match)

        return matches



from __future__ import absolute_import

import os.path

# Use the built-in version of walk if possible, otherwise
# use the scandir module version
try:
    from os import walk
except ImportError:
    from scandir import walk

from glob import glob
from ceres import CeresTree, CeresNode
from django.conf import settings
from graphite.node import BranchNode, LeafNode
from graphite.readers import CeresReader
from graphite.finders.utils import BaseFinder
from graphite.finders import get_real_metric_path, extract_variants
from graphite.tags.utils import TaggedSeries


class CeresFinder(BaseFinder):
    def __init__(self, directory=None):
        directory = directory or settings.CERES_DIR
        self.directory = directory
        self.tree = CeresTree(directory)

    def find_nodes(self, query):
        # translate query pattern if it is tagged
        tagged = not query.pattern.startswith('_tagged.') and ';' in query.pattern
        if tagged:
            # tagged series are stored in ceres using encoded names, so to retrieve them we need to
            # encode the query pattern using the same scheme used in carbon when they are written.
            variants = [
                TaggedSeries.encode(query.pattern, hash_only=True),
                TaggedSeries.encode(query.pattern, hash_only=False),
            ]
        else:
            variants = extract_variants(query.pattern)

        for variant in variants:
            for fs_path in glob(self.tree.getFilesystemPath(variant)):
                metric_path = self.tree.getNodePath(fs_path)

                if CeresNode.isNodeDir(fs_path):
                    ceres_node = self.tree.getNode(metric_path)

                    if ceres_node.hasDataForInterval(query.startTime, query.endTime):
                        real_metric_path = get_real_metric_path(fs_path, metric_path)
                        reader = CeresReader(ceres_node, real_metric_path)
                        # if we're finding by tag, return the proper metric path
                        if tagged:
                            metric_path = query.pattern
                        yield LeafNode(metric_path, reader)

                elif os.path.isdir(fs_path):
                    yield BranchNode(metric_path)

    def get_index(self, requestContext):
        matches = []

        for root, _, files in walk(settings.CERES_DIR):
            root = root.replace(settings.CERES_DIR, '')
            for filename in files:
                if filename == '.ceres-node':
                    matches.append(root)

        return sorted([
            m.replace('/', '.').lstrip('.')
            for m in matches
        ])



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from django.contrib.auth import authenticate, login, logout
try:
    from django.urls import reverse
except ImportError:  # Django < 1.10
    from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import render
from graphite.user_util import getProfile, isAuthenticated


def loginView(request):
    username = request.POST.get('username')
    password = request.POST.get('password')
    if request.method == 'GET':
        nextPage = request.GET.get('nextPage', reverse('browser'))
    else:
        nextPage = request.POST.get('nextPage', reverse('browser'))
    if username and password:
        user = authenticate(username=username,password=password)
        if user is None:
            return render(request, "login.html",{'authenticationFailed' : True, 'nextPage' : nextPage})
        elif not user.is_active:
            return render(request, "login.html",{'accountDisabled' : True, 'nextPage' : nextPage})
        else:
            login(request,user)
            return HttpResponseRedirect(nextPage)
    else:
        return render(request, "login.html",{'nextPage' : nextPage})


def logoutView(request):
    nextPage = request.GET.get('nextPage', reverse('browser'))
    logout(request)
    return HttpResponseRedirect(nextPage)


def editProfile(request):
    if not isAuthenticated(request.user):
        return HttpResponseRedirect(reverse('browser'))
    context = { 'profile' : getProfile(request) }
    return render(request, "editProfile.html",context)


def updateProfile(request):
    profile = getProfile(request,allowDefault=False)
    if profile:
        profile.advancedUI = request.POST.get('advancedUI','off') == 'on'
        profile.save()
    nextPage = request.POST.get('nextPage', reverse('browser'))
    return HttpResponseRedirect(nextPage)
